from __future__ import annotations

from typing import Dict, Tuple

import ada
from ada.visualize.config import ExportConfig


def generate_meta(part: ada.Part, export_config: ExportConfig, sub_geometries: Dict[str, Tuple[str, str]] = None):
    meta = dict()
    for obj in part.get_all_physical_objects(
        sub_elements_only=False,
        filter_by_guids=export_config.data_filter.filter_elements_by_guid,
    ):
        meta[obj.guid] = (obj.name, obj.parent.guid)
        if export_config.data_filter.name_filter is not None and len(export_config.data_filter.name_filter) > 0:
            if obj.name not in [fi.lower() for fi in export_config.data_filter.name_filter]:
                continue
    # Add additional refs to sub-geometry elements
    for key, value in sub_geometries.items():
        meta[key] = value

    for p in part.get_all_parts_in_assembly(True):
        parent_id = p.parent.guid if p.parent is not None else None
        if isinstance(p.parent, ada.Assembly):
            parent_id = "*"
        meta[p.guid] = (p.name, parent_id)

    return meta



from ada.concepts.groups import Group


def write_group(group: Group):
    from ada.ifc.utils import create_guid

    a = group.parent.get_assembly()
    owner_history = a.ifc_store.owner_history
    f = a.ifc_store.f

    ifc_group = f.create_entity("IfcGroup", group.guid, owner_history, group.name, group.description)

    relating_objects = []
    for m in group.members:
        relating_objects.append(a.ifc_store.get_by_guid(m.guid))

    f.create_entity(
        "IfcRelAssignsToGroup",
        create_guid(),
        owner_history,
        group.name,
        group.description,
        RelatedObjects=relating_objects,
        RelatingGroup=ifc_group,
    )



from __future__ import annotations

import xml.etree.ElementTree as ET


def add_local_system(xdir, ydir, zdir) -> ET.Element:
    local_system_elem = ET.Element("local_system")
    d = ["x", "y", "z"]
    for j, vec in enumerate([xdir, ydir, zdir]):
        props = {d[i]: str(k) for i, k in enumerate(vec)}
        props.update(dict(dir=d[j]))
        vec_elem = ET.Element("vector", props)
        local_system_elem.append(vec_elem)

    return local_system_elem



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ada.fem.formats.ifc.writer import to_ifc_fem
from ada.ifc.utils import create_guid, create_local_placement, write_elem_property_sets

if TYPE_CHECKING:
    from ada import Part
    from ada.ifc.store import IfcStore


def write_ifc_spatial_hierarchy(ifc_store: IfcStore):
    sw = SpatialWriter(ifc_store)
    sw.create_ifc_site()


def write_ifc_part(ifc_store: IfcStore, part: Part, include_fem):
    sw = SpatialWriter(ifc_store)

    ifc_part = sw.create_ifc_part(part)

    if len(part.fem.nodes) > 0 and include_fem is True:
        to_ifc_fem(part.fem, ifc_store.f)

    return ifc_part


@dataclass
class SpatialWriter:
    ifc_store: IfcStore

    def create_ifc_site(self):
        assembly = self.ifc_store.assembly

        f = self.ifc_store.f
        owner_history = self.ifc_store.owner_history

        site_placement = create_local_placement(f)
        site = f.create_entity(
            "IfcSite",
            GlobalId=assembly.guid,
            OwnerHistory=owner_history,
            Name=assembly.name,
            Description=None,
            ObjectPlacement=site_placement,
            CompositionType="ELEMENT",
        )
        f.create_entity(
            "IfcRelAggregates",
            create_guid(),
            owner_history,
            "Project Container",
            None,
            f.by_type("IfcProject")[0],
            [site],
        )

        write_elem_property_sets(assembly.metadata, site, f, owner_history)

        return site

    def create_ifc_part(self, part: Part):
        if part.parent is None:
            raise ValueError("Cannot build ifc element without parent")
        from ada.base.ifc_types import SpatialTypes as ITyp

        f = self.ifc_store.f

        owner_history = self.ifc_store.owner_history
        parent = self.ifc_store.get_by_guid(part.parent.guid)

        placement = create_local_placement(
            f,
            origin=part.placement.origin,
            loc_x=part.placement.xdir,
            loc_z=part.placement.zdir,
            relative_to=parent.ObjectPlacement,
        )

        props = dict(
            GlobalId=part.guid,
            OwnerHistory=owner_history,
            Name=part.name,
            Description=part.metadata.get("Description", None),
            ObjectType=None,
            ObjectPlacement=placement,
            Representation=None,
        )

        if part.ifc_class not in (ITyp.IfcElementAssembly, ITyp.IfcGrid):
            props["LongName"] = part.metadata.get("LongName", None)

        if part.ifc_class not in (ITyp.IfcSpatialZone, ITyp.IfcElementAssembly, ITyp.IfcGrid):
            props["CompositionType"] = part.metadata.get("CompositionType", "ELEMENT")

        if part.ifc_class == ITyp.IfcBuildingStorey:
            props["Elevation"] = float(part.placement.origin[2])

        if part.ifc_class == ITyp.IfcGrid:
            from ada.ifc.utils import create_ifcpolyline

            # Todo: Create an IfcGrid object
            props["UAxes"] = [
                f.create_entity(
                    "IfcGridAxis",
                    AxisTag="XAxis",
                    AxisCurve=create_ifcpolyline(f, [(0, 0, 0), (1, 0, 0)]),
                    SameSense=True,
                )
            ]
            props["VAxes"] = [
                f.create_entity(
                    "IfcGridAxis",
                    AxisTag="YAxis",
                    AxisCurve=create_ifcpolyline(f, [(0, 0, 0), (0, 1, 0)]),
                    SameSense=True,
                )
            ]

        ifc_elem = f.create_entity(part.ifc_class.value, **props)

        existing_rel_agg = False
        for rel_agg in f.by_type("IfcRelAggregates"):
            if rel_agg.RelatingObject == parent:
                rel_agg.RelatedObjects = tuple([*rel_agg.RelatedObjects, ifc_elem])
                existing_rel_agg = True
                break

        if existing_rel_agg is False:
            f.create_entity(
                "IfcRelAggregates",
                GlobalId=create_guid(),
                OwnerHistory=owner_history,
                Name="Site Container",
                Description=None,
                RelatingObject=parent,
                RelatedObjects=[ifc_elem],
            )

        write_elem_property_sets(part.metadata, ifc_elem, f, owner_history)

        return ifc_elem



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, List, Union

import numpy as np

from ada.concepts.points import Node
from ada.config import get_logger

from .common import Csys, FemBase
from .shapes import ElemShape, ElemType
from .shapes import definitions as shape_def
from .shapes.definitions import LineShapes, ShapeResolver, ShellShapes, SolidShapes

if TYPE_CHECKING:
    from ada import FEM, Beam, Pipe, Plate, Shape, Wall
    from ada.fem import ConnectorSection, FemSection, FemSet

logger = get_logger()


class Elem(FemBase):
    EL_TYPES = ElemType

    def __init__(
        self,
        el_id,
        nodes: list[Node],
        el_type: str | ShellShapes | LineShapes | SolidShapes,
        elset=None,
        fem_sec: FemSection = None,
        mass_props=None,
        parent: FEM = None,
        el_formulation_override=None,
        metadata=None,
    ):
        super(Elem, self).__init__(el_id, metadata, parent)
        self.type = el_type
        self._el_id = el_id
        self._shape = None

        if nodes is not None and isinstance(nodes[0], Node):
            for node in nodes:
                node.add_obj_to_refs(self)

        self._nodes = nodes
        self._elset = elset
        self._fem_sec = fem_sec
        self._mass_props = mass_props
        self._hinge_prop = None
        self._eccentricity = None
        self._formulation_override = el_formulation_override
        self._refs = []

    def get_offset_coords(self):
        nodes = [n.p for n in self.nodes]
        if self.eccentricity is None:
            return nodes

        mat = np.eye(3)

        e1 = self.eccentricity.end1
        e2 = self.eccentricity.end2

        if e1 is not None:
            ecc1 = e1.ecc_vector
            n_old = e1.node
            nodes[0] = np.dot(mat, ecc1) + n_old.p

        if e2 is not None:
            ecc2 = e2.ecc_vector
            n_old = e2.node
            nodes[-1] = np.dot(mat, ecc2) + n_old.p

        return nodes

    def replace_node_with_other_node(self, old_node: Node, new_node: Node):
        index = None
        for i, node in enumerate(self.nodes):
            if node == old_node:
                index = i
        if index is None:
            raise ValueError(f'Unable to find {old_node.id} in this element "{self.id}"')
        self.nodes.pop(index)
        self.nodes.insert(index, new_node)

    @property
    def type(self) -> ShellShapes | LineShapes | SolidShapes:
        return self._el_type

    @type.setter
    def type(self, value: str | ShellShapes | LineShapes | SolidShapes):
        if isinstance(value, str):
            result = ShapeResolver.get_el_type_from_str(value)
            if result is None:
                raise ValueError(f'Currently unsupported element type "{value}".')
        else:
            result = value

        self._el_type = result

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def id(self) -> int:
        return self._el_id

    @id.setter
    def id(self, value):
        if type(value) not in (np.int32, int, np.uint64, np.int64) and issubclass(type(self), Connector) is False:
            raise ValueError(f'Element ID "{type(value)}" must be numeric')
        self._el_id = value

    @property
    def nodes(self) -> List[Node]:
        return self._nodes

    @property
    def hinge_prop(self) -> Union[None, HingeProp]:
        return self._hinge_prop

    @hinge_prop.setter
    def hinge_prop(self, value: HingeProp):
        self._hinge_prop = value

    @property
    def eccentricity(self) -> Union[None, Eccentricity]:
        return self._eccentricity

    @eccentricity.setter
    def eccentricity(self, value: Eccentricity):
        self._eccentricity = value

    @property
    def elset(self):
        return self._elset

    @elset.setter
    def elset(self, value: FemSet):
        self._elset = value

    @property
    def fem_sec(self) -> FemSection:
        return self._fem_sec

    @fem_sec.setter
    def fem_sec(self, value: FemSection):
        self._fem_sec = value

    @property
    def mass_props(self) -> Mass:
        return self._mass_props

    @mass_props.setter
    def mass_props(self, value: Mass):
        self._mass_props = value

    @property
    def shape(self) -> ElemShape:
        if self._shape is None:
            self._shape = ElemShape(self.type, self.nodes)
        return self._shape

    @property
    def refs(self) -> List[Union[Elem, Beam, Plate, Pipe, Wall, Shape]]:
        return self._refs

    @property
    def formulation_override(self):
        return self._formulation_override if self._formulation_override is not None else self.type

    def update(self) -> None:
        self._nodes = list(set(self.nodes))
        if len(self.nodes) <= 1:
            self._el_id = None
        else:
            self._shape = None

    def updating_nodes(self, old_node: Node, new_node: Node) -> None:
        """Exchanging old node with new node, and updating the element shape"""
        node_index = self.nodes.index(old_node)
        self.nodes.pop(node_index)
        self.nodes.insert(node_index, new_node)

        self.update()

    def __repr__(self):
        nodes = self.nodes if hasattr(self, "_nodes") else "Nodes not yet initialized"
        return f'Elem(ID: {self._el_id}, Type: {self.type}, NodeIds: "{nodes}")'


@dataclass
class Hinge:
    retained_dofs: List[int]
    csys: Csys
    concept_node: Node = None
    fem_node: Node = None
    elem_n_index: int = None
    beam_n_index: int = None
    constraint_ref = None


@dataclass
class HingeProp:
    end1: Hinge = None
    end2: Hinge = None
    elem_ref: Elem = None
    beam_ref: Beam = None


@dataclass
class EccPoint:
    node: Node
    ecc_vector: np.ndarray


@dataclass
class Eccentricity:
    end1: EccPoint = None
    end2: EccPoint = None
    sh_ecc_vector: np.ndarray = None


class ConnectorTypes:
    BUSHING = "bushing"
    CARTESIAN = "cartesian"

    all = [BUSHING, CARTESIAN]


class Connector(Elem):
    CON_TYPES = ConnectorTypes

    def __init__(
        self,
        name,
        el_id,
        n1: Node,
        n2: Node,
        con_type,
        con_sec: "ConnectorSection",
        preload=None,
        csys: Csys = None,
        metadata=None,
        parent: "FEM" = None,
    ):
        if type(n1) is not Node or type(n2) is not Node:
            raise ValueError("Connector Start\\end must be nodes")
        super(Connector, self).__init__(el_id, [n1, n2], ElemType.CONNECTOR_SHAPES.CONNECTOR)
        super(Elem, self).__init__(name, metadata, parent)
        self._n1 = n1
        self._n2 = n2
        self._con_type = con_type
        self._con_sec = con_sec
        self._preload = preload
        self._csys = csys if csys is not None else Csys(f"{name}_csys")

    @property
    def con_type(self):
        return self._con_type

    @con_type.setter
    def con_type(self, value: str):
        self._con_type = value

    @property
    def con_sec(self) -> "ConnectorSection":
        return self._con_sec

    @con_sec.setter
    def con_sec(self, value: "ConnectorSection"):
        self._con_sec = value

    @property
    def n1(self) -> Node:
        return self._n1

    @n1.setter
    def n1(self, value: Node):
        self._n1 = value

    @property
    def n2(self) -> Node:
        return self._n2

    @n2.setter
    def n2(self, value: Node):
        self._n2 = value

    @property
    def csys(self) -> Csys:
        return self._csys

    @csys.setter
    def csys(self, value: Csys):
        self._csys = value

    def __repr__(self):
        return f'ConnectorElem(ID: {self.id}, Type: {self.type}, End1: "{self.n1}", End2: "{self.n2}")'


class Spring(Elem):
    def __init__(self, name, el_id, el_type, stiff, fem_set: "FemSet", metadata=None, parent=None):
        super(Spring, self).__init__(el_id, fem_set.members, el_type)
        super(Elem, self).__init__(name, metadata, parent)
        self._stiff = stiff
        self._n1 = fem_set.members[0]
        self._n2 = None
        if len(fem_set.members) > 1:
            self._n2 = fem_set.members[-1]
        self._fem_set = fem_set

    @property
    def fem_set(self) -> "FemSet":
        return self._fem_set

    @property
    def stiff(self):
        return self._stiff

    def __repr__(self):
        return f'Spring("{self.name}", type="{self._stiff}")'


class MassTypes:
    MASS = shape_def.MassTypes.MASS
    NONSTRU = "NONSTRUCTURAL MASS"
    ROT_INERTIA = shape_def.MassTypes.ROTARYI

    all = [MASS, NONSTRU, ROT_INERTIA]


class MassPType:
    ISOTROPIC = "ISOTROPIC"
    ANISOTROPIC = "ANISOTROPIC"

    all = [ISOTROPIC, ANISOTROPIC]


class Mass(Elem):
    TYPES = MassTypes
    PTYPES = MassPType

    def __init__(
        self,
        name,
        ref: FemSet | list[Node] | None,
        mass,
        mass_type=None,
        ptype=None,
        mass_id: int = None,
        units=None,
        metadata=None,
        parent=None,
    ):
        if hasattr(ref, "members"):
            self._fem_set = ref
            members = ref.members
        else:
            members = ref

        if mass is None:
            raise ValueError("Mass cannot be None")

        if type(mass) not in (list, tuple):
            logger.info(f"Mass {type(mass)} converted to list of len=1. Assume equal mass in get_all 3 transl. DOFs.")
            ptype = self.PTYPES.ISOTROPIC
            mass = [mass]

        self._mass = mass
        if isinstance(mass_type, str):
            mass_type = shape_def.MassTypes.from_str(mass_type)
            if mass_type is None:
                raise ValueError(f'Mass type "{self.type}" is not {shape_def.MassTypes}')

        elif mass_type is None:
            mass_type = self.TYPES.MASS

        self._el_type = mass_type

        if ptype not in MassPType.all and ptype is not None:
            raise ValueError(f'Mass point type "{ptype}" is not in list of supported types {MassPType.all}')

        super(Mass, self).__init__(mass_id, members, self.type)
        super(Elem, self).__init__(name, metadata, parent)

        self.point_mass_type = ptype
        self._units = units
        self._members = members
        self._elset = None
        self._check_input()

    def _check_input(self):
        if self.point_mass_type is None:
            if self.type == MassTypes.MASS:
                if type(self._mass) in (list, tuple):
                    raise ValueError("Mass can only be a scalar number for Isotropic mass")
        elif self.point_mass_type == MassPType.ISOTROPIC:
            if (len(self._mass) == 1) is False:
                raise ValueError("Mass can only be a scalar number for Isotropic mass")
        elif self.point_mass_type == MassPType.ANISOTROPIC:
            if (len(self._mass) == 3) is False:
                raise ValueError("Mass must be specified for 3 dofs for Anisotropic mass")
        else:
            raise ValueError(f'Unknown mass input "{self.type}"')

    @property
    def fem_set(self) -> FemSet:
        return self._fem_set

    @fem_set.setter
    def fem_set(self, value: FemSet):
        self._members = value.members
        self._fem_set = value

    @property
    def elset(self):
        return self._elset

    @elset.setter
    def elset(self, value):
        self._elset = value

    @property
    def mass(self):
        if self.point_mass_type is None:
            if self.type == MassTypes.MASS:
                if type(self._mass) in (list, tuple):
                    raise ValueError("Mass can only be a scalar number for Isotropic mass")
                return float(self._mass[0])
            elif self.type == MassTypes.NONSTRU:
                return self._mass
            elif self.type == MassTypes.ROT_INERTIA:
                return self._mass
            else:
                return float(self._mass)
        elif self.point_mass_type == MassPType.ISOTROPIC:
            if (len(self._mass) == 1) is False:
                raise ValueError("Mass can only be a scalar number for Isotropic mass")
            return self._mass[0]
        elif self.point_mass_type == MassPType.ANISOTROPIC:
            if (len(self._mass) == 3) is False:
                raise ValueError("Mass must be specified for 3 dofs for Anisotropic mass")
            return self._mass
        else:
            raise ValueError(f'Unknown mass input "{self.type}"')

    @mass.setter
    def mass(self, value) -> None:
        self._mass = value

    @property
    def members(self):
        return self._members

    @property
    def units(self):
        return self._units

    @property
    def point_mass_type(self):
        return self._ptype

    @point_mass_type.setter
    def point_mass_type(self, value):
        self._ptype = value

    def __repr__(self) -> str:
        return f"Mass({self.name}, {self.point_mass_type}, [{self.mass}])"


def find_element_type_from_list(elements: List[Elem]) -> str:
    el_types = set(el.shape.elem_type_group for el in elements)
    if len(el_types) != 1:
        raise NotImplementedError("Mixed element set types as basis for surface sets is not yet supported")
    return elements[0].shape.elem_type_group



from __future__ import annotations

import xml.etree.ElementTree as ET
from typing import TYPE_CHECKING

from .write_utils import add_local_system

if TYPE_CHECKING:
    from ada import Beam, Part


def add_beams(root: ET.Element, part: Part, sat_map: dict):
    from ada import Beam

    for beam in part.get_all_physical_objects(by_type=Beam):
        add_straight_beam(beam, root)


def add_straight_beam(beam: Beam, xml_root: ET.Element):
    structure_elem = ET.Element("structure")
    beam_elem = ET.Element("straight_beam", {"name": beam.name})
    structure_elem.append(beam_elem)
    beam_elem.append(add_local_system(beam.xvec, beam.yvec, beam.up))
    beam_elem.append(add_segments(beam))
    xml_root.append(structure_elem)


def add_segments(beam: Beam):
    segments = ET.Element("segments")
    props = dict(index="1", section_ref=beam.section.name, material_ref=beam.material.name)
    straight_segment = ET.SubElement(segments, "straight_segment", props)

    d = ["x", "y", "z"]

    geom = ET.SubElement(straight_segment, "geometry")
    wire = ET.SubElement(geom, "wire")
    guide = ET.SubElement(wire, "guide")
    for i, pos in enumerate([beam.n1, beam.n2], start=1):
        props = {d[i]: str(k) for i, k in enumerate(pos.p)}
        props.update(dict(end=str(i)))
        ET.SubElement(guide, "position", props)

    ET.SubElement(wire, "sat_reference")

    # TODO: add SAT embedded geometry and include the reference to the EDGE geometry here
    # ET.SubElement(sat_ref, "edge_ref", dict(edge_ref=""))

    return segments



from typing import TYPE_CHECKING

from ada.fem import PredefinedField

from .helper_utils import get_instance_name

if TYPE_CHECKING:
    from ada import FEM


def predefined_fields_str(fem: "FEM"):
    def eval_fields(pre_field: PredefinedField):
        return True if pre_field.type != PredefinedField.TYPES.INITIAL_STATE else False

    return "\n".join(
        [predefined_field_str(prefield) for prefield in filter(eval_fields, fem.predefined_fields.values())]
    )


def predefined_field_str(pre_field: PredefinedField) -> str:
    dofs_str = ""
    for dof, magn in zip(pre_field.dofs, pre_field.magnitude):
        if float(magn) == 0.0:
            continue
        dofs_str += f"{get_instance_name(pre_field.fem_set, True)}, {dof}, {magn}\n"
    dofs_str.rstrip()
    return f"""** PREDEFINED FIELDS
**
** Name: {pre_field.name}   Type: {pre_field.type}
*Initial Conditions, type={pre_field.type}
{dofs_str}"""



from typing import TYPE_CHECKING

from ada.fem import Amplitude

if TYPE_CHECKING:
    from ada import FEM


def amplitudes_str(fem: "FEM"):
    return "\n".join([amplitude_str(ampl) for ampl in fem.amplitudes.values()])


def amplitude_str(amplitude: Amplitude) -> str:
    name, x, y, smooth = amplitude.name, amplitude.x, amplitude.y, amplitude.smooth
    a = 1
    data = ""
    for i, var in enumerate(zip(list(x), list(y))):
        if a == 4:
            if i == len(list(x)) - 1:
                data += "{:.4E}, {:.4E}, ".format(var[0], var[1])
            else:
                data += "{:.4E}, {:.4E},\n         ".format(var[0], var[1])
            a = 0
        else:
            data += "{:.4E}, {:.4E}, ".format(var[0], var[1])
        a += 1

    smooth = ", DEFINITION=TABULAR, SMOOTH={}".format(smooth) if smooth is not None else ""
    amplitude = """*Amplitude, name={0}{2}\n         {1}\n""".format(name, data, smooth)
    return amplitude.rstrip()



from __future__ import annotations

from typing import TYPE_CHECKING

from ada.fem import StepEigen

if TYPE_CHECKING:
    from ada.concepts.spatial import Part


def step_eig_str(step: StepEigen, part: Part) -> str:
    bcs = part.fem.bcs

    assembly = part.get_assembly()
    if part != assembly:
        for bc in assembly.fem.bcs:
            if bc not in bcs:
                bcs.append(bc)

    if len(bcs) > 1 or len(bcs) == 0:
        raise NotImplementedError(f"Number of BC sets {len(bcs)=} is for now limited to 1 for eigenfrequency analysis")

    eig_map = dict(sorensen="SORENSEN", lanczos="TRI_DIAG")
    eig_type = step.metadata.get("eig_method", "sorensen")
    eig_method = eig_map[eig_type]

    sec_str = ""
    if len(part.fem.sections.lines) > 0 or len(part.fem.sections.shells) > 0:
        sec_str = "\n    CARA_ELEM=element,"

    # TODO: Add check for second order shell elements. If exists add conversion of results back from TRI7 to TRI6
    _ = """
    model_0 = AFFE_MODELE(
    AFFE=(
        _F(GROUP_MA=sh_2nd_order_sets, PHENOMENE='MECANIQUE', MODELISATION='MEMBRANE',),
    ),
    MAILLAGE=mesh
)

modes_0 = PROJ_CHAMP(
    MODELE_1=model,
    MODELE_2=model_0,
    RESULTAT=modes
)"""

    bc = bcs[0]
    return f"""
#modal analysis
ASSEMBLAGE(
    MODELE=model,
    CHAM_MATER=material,{sec_str}
    CHARGE={bc.name},
    NUME_DDL=CO('dofs_eig'),
    MATR_ASSE = (
        _F(MATRICE=CO('stiff'), OPTION ='RIGI_MECA',),
        _F(MATRICE=CO('mass'), OPTION ='MASS_MECA', ),
    ),
)
# Using Subspace Iteration method ('SORENSEN' AND 'PLUS_PETITE')
# See https://www.code-aster.org/V2/UPLOAD/DOC/Formations/01-modal-analysis.pdf for more information
#

modes = CALC_MODES(
    CALC_FREQ=_F(NMAX_FREQ={step.num_eigen_modes}, ) ,
    SOLVEUR_MODAL=_F(METHODE='{eig_method}'),
    MATR_MASS=mass,
    MATR_RIGI=stiff,
    OPTION='PLUS_PETITE',
    VERI_MODE=_F(STOP_ERREUR='NON')
)



IMPR_RESU(
    RESU=_F(RESULTAT=modes, TOUT_CHAM='OUI'),
    UNITE=80
)
"""



from ada import Material, Part
from ada.concepts.containers import Materials
from ada.materials.metals import Metal


def get_materials(xml_root, parent) -> Materials:
    all_mats = xml_root.findall(".//material")
    materials = [interpret_material(mat_el.attrib["name"], mat_el[0], parent) for mat_el in all_mats]
    return Materials(materials, parent)


def interpret_material(name, mat_prop, parent: Part):
    mat_prop_map = dict(
        isotropic_linear_material=isotropic_linear_material, isotropic_shear_material=isotropic_shear_material
    )
    mat_interpreter = mat_prop_map.get(mat_prop.tag, None)
    if mat_interpreter is None:
        raise ValueError(f"Missing property {mat_prop.tag}")

    return mat_interpreter(name, mat_prop, parent)


def isotropic_linear_material(name, mat_prop, parent: Part) -> Material:
    model = Metal(
        sig_y=float(mat_prop.attrib["yield_stress"]),
        rho=float(mat_prop.attrib["density"]),
        E=float(mat_prop.attrib["youngs_modulus"]),
        v=float(mat_prop.attrib["poissons_ratio"]),
        alpha=float(mat_prop.attrib["thermal_expansion"]),
        zeta=float(mat_prop.attrib["damping"]),
        sig_u=None,
        plasticitymodel=None,
    )
    return Material(name=name, mat_model=model, parent=parent)


def isotropic_shear_material(name, mat_prop, parent: Part) -> Material:
    model = Metal(
        sig_y=355e6,
        rho=float(mat_prop.attrib["density"]),
        E=float(mat_prop.attrib["youngs_modulus"]),
        v=float(mat_prop.attrib["poissons_ratio"]),
        alpha=float(mat_prop.attrib["thermal_expansion"]),
        zeta=float(mat_prop.attrib["damping"]),
        sig_u=None,
        plasticitymodel=None,
    )
    return Material(name=name, mat_model=model, parent=parent)



from typing import TYPE_CHECKING

import numpy as np

from .vector_utils import vector_length

if TYPE_CHECKING:
    from ada import Beam, Plate


def align_to_plate(plate: "Plate"):
    normal = plate.poly.normal
    h = plate.t * 5
    origin = plate.poly.placement.origin - h * normal * 1.1 / 2
    xdir = plate.poly.placement.xdir
    return dict(h=h, normal=normal, origin=origin, xdir=xdir)


def align_to_beam(beam: "Beam"):
    ymin = beam.yvec * np.array(beam.bbox().p1)
    ymax = beam.yvec * np.array(beam.bbox().p2)
    origin = beam.n1.p - ymin * 1.1
    normal = -beam.yvec
    xdir = beam.xvec
    h = vector_length(ymax - ymin) * 1.2
    return dict(h=h, normal=normal, origin=origin, xdir=xdir)



from __future__ import annotations

from itertools import groupby
from operator import attrgetter
from typing import TYPE_CHECKING, Iterable

from ada.core.utils import NewLine
from ada.fem.shapes import definitions as shape_def

from .helper_utils import get_instance_name
from .write_masses import write_mass_elem

if TYPE_CHECKING:
    from ada import FEM
    from ada.fem import Elem, FemSet


def elements_str(fem: "FEM", written_on_assembly_level: bool) -> str:
    part_el = fem.elements
    grouping = groupby(part_el, key=attrgetter("type", "elset"))
    if len(fem.elements) == 0:
        return "** No elements"

    return "".join(
        list(
            filter(
                lambda x: x is not None,
                [elwriter(x, elements, fem, written_on_assembly_level) for x, elements in grouping],
            )
        )
    ).rstrip()


def write_elements(
    eltype: shape_def.LineShapes | shape_def.ShellShapes | shape_def.SolidShapes,
    elset: FemSet,
    fem: FEM,
    elements: Iterable[Elem],
    alevel: bool,
):
    el_type = fem.options.ABAQUS.default_elements.get_element_type(eltype)
    el_set_str = f", ELSET={elset.name}" if elset is not None else ""
    el_str = "\n".join((write_elem(el, alevel) for el in elements))
    return f"""*ELEMENT, type={el_type}{el_set_str}\n{el_str}\n"""


def write_elem(el: Elem, alevel: bool) -> str:
    nl = NewLine(10, suffix=7 * " ")
    if len(el.nodes) > 6:
        di = " {}"
    else:
        di = "{:>13}"
    el_str = (
        f"{el.id:>7}, " + " ".join([f"{di.format(get_instance_name(no, alevel))}," + next(nl) for no in el.nodes])[:-1]
    )
    return el_str


def elwriter(eltype_set, elements, fem: "FEM", written_on_assembly_level: bool):
    eltype, elset = eltype_set
    if isinstance(eltype, shape_def.ConnectorTypes):
        return None
    elif isinstance(eltype, shape_def.MassTypes):
        return write_mass_elem(eltype, elset, fem, elements, written_on_assembly_level)
    else:
        return write_elements(eltype, elset, fem, elements, written_on_assembly_level)



from __future__ import annotations

import os
import pathlib
from typing import TYPE_CHECKING, List

from ada.config import get_logger
from ada.fem import StepEigen
from ada.fem.exceptions.fea_execution import (
    FEAnalysisUnableToStart,
    FEAnalysisUnsuccessfulError,
)
from ada.fem.formats.utils import DatFormatReader

from .read_odb import get_odb_data

if TYPE_CHECKING:
    from ada.fem.results.concepts import ElementDataOutput, FEMDataOutput, Results
    from ada.fem.results.eigenvalue import EigenDataSummary

logger = get_logger()


def get_eigen_data(dat_file: str | os.PathLike) -> EigenDataSummary:
    from ada.fem.results.eigenvalue import EigenDataSummary, EigenMode

    dtr = DatFormatReader()

    re_compiled = dtr.compile_ff_re([int] + [float] * 5)
    re_compiled_2 = dtr.compile_ff_re([int] + [float] * 6)

    eig_str = "eigenvalueoutput"
    part_str = "participationfactors"
    eff_modal = "effectivemass"

    eig_res = dtr.read_data_lines(dat_file, re_compiled, eig_str, part_str, split_data=True)
    part_res = dtr.read_data_lines(dat_file, re_compiled_2, part_str, eff_modal, split_data=True)
    modalmass = dtr.read_data_lines(dat_file, re_compiled_2, eff_modal, split_data=True)

    eigen_modes: List[EigenMode] = []

    dof_base = ["x", "y", "z", "rx", "ry", "rz"]
    part_factor_names = ["p" + x for x in dof_base]
    eff_mass_names = ["ef" + x for x in dof_base]

    # Note! participation factors and effective modal mass are each deconstructed into 6 degrees of freedom
    for eig, part, modal in zip(eig_res, part_res, modalmass):
        mode, eig_value, freq_rad, freq_cycl, gen_mass, composite_modal_damping = eig
        eig_output = dict(eigenvalue=eig_value, f_rad=freq_rad, f_hz=freq_cycl)
        participation_data = {pn: p for pn, p in zip(part_factor_names, part[1:])}
        eff_mass_data = {pn: p for pn, p in zip(eff_mass_names, part[1:])}
        eigen_modes.append(EigenMode(no=mode, **eig_output, **participation_data, **eff_mass_data))

    return EigenDataSummary(eigen_modes)


def read_abaqus_results(results: "Results", file_ref: pathlib.Path, overwrite):
    dat_file = file_ref.with_suffix(".dat")
    if results.assembly is not None and results.assembly.fem.steps[0] == StepEigen:
        # TODO: Figure out if it is worthwhile adding support for reading step information or if it should be explicitly
        #   stated
        pass

    if dat_file.exists():
        results.eigen_mode_data = get_eigen_data(dat_file)

    check_execution(file_ref)

    logger.error("Result mesh data extraction is not supported for abaqus")

    return odb_data_to_results(file_ref, results)


def check_execution(file_ref: pathlib.Path):
    sta_file = file_ref.with_suffix(".sta")
    if sta_file.exists() is False:
        raise FEAnalysisUnableToStart()

    with open(sta_file, "r") as f:
        if "THE ANALYSIS HAS NOT BEEN COMPLETED" in f.read():
            raise FEAnalysisUnsuccessfulError()


def odb_data_to_results(odb_file: pathlib.Path, results: Results) -> None:
    from ada.fem.results.concepts import HistoryStepDataOutput, ResultsHistoryOutput

    odb_data = get_odb_data(odb_file)
    res = ResultsHistoryOutput()

    for step in odb_data["steps"].values():
        name = step["name"]
        step_type = step["procedure"]
        step_res = HistoryStepDataOutput(name=name, step_type=step_type)
        res.steps.append(step_res)

        for reg in step["historyRegions"].values():
            history_outputs = reg["historyOutputs"].values()
            name = reg["name"]
            if "element" in name.lower():
                step_res.element_data[name] = get_element_component_data(name, history_outputs)
            else:
                step_res.fem_data = get_fem_data_output(history_outputs)

    results.history_output = res


def get_element_component_data(name: str, history_outputs: dict) -> ElementDataOutput:
    from ada.fem.results.concepts import ElementDataOutput, ElemForceComp

    cu_map = {"CU1": 0, "CU2": 1, "CU3": 2, "CUR1": 3, "CUR2": 4, "CUR3": 5}
    cf_map = {"CTF1": 0, "CTF2": 1, "CTF3": 2, "CTM1": 3, "CTM2": 4, "CTM3": 5}
    displ_data = dict()
    force_data = dict()
    for data in history_outputs:
        comp = data["name"]
        cu = cu_map.get(comp, None)
        cf = cf_map.get(comp, None)
        if cu is not None:
            displ_data[cu] = [tuple(x) for x in data["data"]]
        elif cf is not None:
            force_data[cf] = ElemForceComp(comp, [tuple(x) for x in data["data"]])

    return ElementDataOutput(name=name, displacements=displ_data, forces=force_data)


def get_fem_data_output(history_outputs) -> dict[str, FEMDataOutput]:
    from ada.fem.results.concepts import FEMDataOutput

    return {x["name"]: FEMDataOutput(x["name"], [tuple(y) for y in x["data"]]) for x in history_outputs}



import json
import os
import pathlib
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ada import Assembly


def to_three_json(assembly: "Assembly", output_file_path):
    from OCC.Core.Tesselator import ShapeTesselator

    quality = 1.0
    render_edges = False
    parallel = True
    total_json = []
    for p in assembly.parts.values():
        for obj in p.get_all_physical_objects():
            geom = obj.solid()
            tess = ShapeTesselator(geom)
            tess.Compute(compute_edges=render_edges, mesh_quality=quality, parallel=parallel)
            res = tess.ExportShapeToThreejsJSONString(obj.name)
            total_json.append(res)

    output = {
        "metadata": {"version": 4.3, "type": "Object", "generator": "ObjectExporter"},
        "textures": [],
        "images": [],
        "geometries": [
            {
                "uuid": "0A8F2988-626F-411C-BD6A-AC656C4E6878",
                "type": "BufferGeometry",
                "data": {
                    "attributes": {
                        "position": {
                            "itemSize": 3,
                            "type": "Float32Array",
                            "array": [1, 1, 0, 1, -1, 0, -1, -1, 0, -1, 1, 0],
                            "normalized": False,
                        },
                        "normal": {
                            "itemSize": 3,
                            "type": "Float32Array",
                            "array": [0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1],
                            "normalized": False,
                        },
                        "uv": {
                            "itemSize": 2,
                            "type": "Float32Array",
                            "array": [1, 1, 1, 0, 0, 0, 0, 1],
                            "normalized": False,
                        },
                    },
                    # // type of index must be Uint8Array or Uint16Array.
                    # // # vertices thus cannot exceed 255 or 65535 respectively.
                    # // The current parser is able to read the index array
                    # // if it is nested in the attributes object, but such
                    # // syntax is no longer encouraged.
                    "index": {"type": "Uint16Array", "array": [0, 1, 2, 0, 2, 3]},
                    "boundingSphere": {"center": [0, 0, 0], "radius": 1},
                },
            }
        ],
        "materials": [],
        "object": {
            "uuid": "378FAA8D-0888-4249-8701-92D1C1F37C51",
            "type": "Scene",
            "matrix": [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1],
            "children": [
                {
                    "uuid": "E7B44C44-DD75-4C29-B571-21AD6AEF0CA9",
                    "name": "SharedVertexTest",
                    "type": "Mesh",
                    "geometry": "0A8F2988-626F-411C-BD6A-AC656C4E6878",
                    "matrix": [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1],
                }
            ],
        },
    }

    output_file_path = pathlib.Path(output_file_path)
    os.makedirs(output_file_path.parent, exist_ok=True)
    with open(output_file_path, "w") as f:
        json.dump(output, f, indent=4)



from __future__ import annotations

from dataclasses import dataclass

import numpy as np


@dataclass
class VisColor:
    name: str
    pbrMetallicRoughness: PbrMetallicRoughness
    used_by: list[str]


@dataclass
class PbrMetallicRoughness:
    baseColorFactor: list[float]
    metallicFactor: float
    roughnessFactor: float


class DataColorizer:
    default_palette = [(0, 149 / 255, 239 / 255), (1, 0, 0)]

    @staticmethod
    def colorize_data(data: np.ndarray, func=None, palette=None):
        if func is None:
            shape = data.shape
            num_cols = shape[1]
            if num_cols in (3, 6):
                func = magnitude
            elif num_cols == 1:
                func = magnitude1d

        palette = DataColorizer.default_palette if palette is None else palette

        res = [func(d) for d in data]
        sorte = sorted(res)
        min_r = sorte[0]
        max_r = sorte[-1]

        start = np.array(palette[0])
        end = np.array(palette[-1])

        def curr_p(t):
            return start + (end - start) * t / (max_r - min_r)

        colors = np.asarray(list([curr_p(x) for x in res]), dtype="float32")
        return colors


def magnitude(u):
    return np.sqrt(u[0] ** 2 + u[1] ** 2 + u[2] ** 2)


def magnitude2d(u):
    return np.sqrt(u[0] ** 2 + u[1] ** 2)


def magnitude1d(u):
    return u



from __future__ import annotations

import json
import os
import pathlib
import re
import shutil
import subprocess
import sys
from contextlib import contextmanager
from itertools import chain
from typing import TYPE_CHECKING

from send2trash import send2trash

from ada.concepts.containers import Beams, Plates
from ada.config import Settings, get_logger
from ada.fem import Elem
from ada.fem.exceptions import FEASolverNotInstalled

if TYPE_CHECKING:
    from ada import Assembly, Beam, Part, Plate
    from ada.fem.formats.general import FEATypes

logger = get_logger()


class DatFormatReader:
    re_flags = re.MULTILINE | re.DOTALL
    re_int = r"[0-9]{1,9}"
    re_decimal_float = r"[+|-]{0,1}[0-9]{1,9}\.[0-9]{0,6}"
    re_decimal_scientific = r"[+|-]{0,1}[0-9]{1,2}\.[0-9]{5,7}[E|e][\+|\-][0-9]{2}"

    def compile_ff_re(self, list_of_types, separator=None):
        """Create a compiled regex pattern for a specific combination of floats and ints provided"""
        re_str = r"^\s*("
        sep_str = "" if separator is None else separator
        for i, t in enumerate(list_of_types):
            if i == len(list_of_types) - 1:
                sep_str = ""
            if t is int:
                re_str += rf"{self.re_int}{sep_str}\s*"
            elif t is float:
                re_str += rf"(?:{self.re_decimal_scientific}|{self.re_decimal_float}){sep_str}\s*"
            else:
                raise ValueError()
        re_str += r")\n"
        return re.compile(re_str, self.re_flags)

    def read_data_lines(self, dat_file, regex: re.Pattern, start_flag, end_flag=None, split_data=False) -> list:
        """Reads line by line without any spaces to search for strings while disregarding formatting"""
        read_data = False
        results = []
        with open(dat_file, "r") as f:
            for line in f.readlines():
                compact_str = line.replace(" ", "").strip().lower()
                if start_flag in compact_str:
                    read_data = True
                if end_flag is not None and end_flag in compact_str:
                    return results
                if read_data is False:
                    continue
                res = regex.search(line)
                if res is not None:
                    result_data = res.group(1)
                    if split_data:
                        result_data = result_data.split()
                    results.append(result_data)

        return results


class LocalExecute:
    """Backend Component for executing local analysis"""

    def __init__(
        self,
        inp_path,
        cpus=2,
        gpus=None,
        run_ext=False,
        metadata=None,
        auto_execute=True,
        excute_locally=True,
        run_in_shell=False,
    ):
        self._inp_path = pathlib.Path(inp_path)
        self._cpus = cpus
        self._gpus = gpus
        self.run_ext = run_ext
        self._metadata = metadata
        self.auto_execute = auto_execute
        self.local_execute = excute_locally
        self.run_in_shell = run_in_shell

    def _run_local(self, run_command, stop_command=None, exit_on_complete=True, bat_start_str=None):
        if self._metadata is not None:
            with open(self.inp_path.parent / "analysis_manifest.json", "w") as fp:
                json.dump(self._metadata, fp, indent=4)

        if sys.platform == "linux" or sys.platform == "linux2":
            logger.info("Running on Linux platform.")
            out = run_linux(self, run_command)
        elif sys.platform == "darwin":
            logger.info("Running on macOS platform.")
            out = run_macOS(self, run_command)
        else:  # sys.platform == "win32":
            logger.info("Running on Windows platform.")
            out = run_windows(
                self,
                run_command,
                stop_command,
                exit_on_complete,
                bat_start_str=bat_start_str,
                run_in_shell=self.run_in_shell,
            )

        return out

    def get_exe(self, fea_software):
        from ada.fem.formats.general import fem_solver_map

        solver_exe_name = fem_solver_map.get(fea_software, fea_software)
        exe_path = None
        for exe_test in [fea_software, solver_exe_name]:
            try:
                exe_path = get_exe_path(exe_test)
            except FileNotFoundError:
                continue
            if exe_path is not None:
                break

        if exe_path is None:
            msg = (
                f'FEA Solver executable for "{solver_exe_name}" is not found. '
                f"Please make sure that an executable exists at the specified location.\n"
                f"See section about adding FEA solvers to paths "
                f"so that adapy finds them in the readme at https://github.com/Krande/adapy"
            )
            raise FEASolverNotInstalled(msg)

        return exe_path

    def run(self):
        raise NotImplementedError("The run function is not implemented")

    @property
    def analysis_dir(self):
        return self.inp_path.parent

    @property
    def execute_dir(self):
        if Settings.execute_dir is None:
            return self.analysis_dir
        else:
            return Settings.execute_dir / self.analysis_name

    @property
    def analysis_name(self):
        return self.inp_path.stem

    @property
    def inp_path(self):
        return pathlib.Path(self._inp_path)

    @inp_path.setter
    def inp_path(self, value):
        self._inp_path = pathlib.Path(value)

    @property
    def cpus(self):
        return self._cpus


def is_buffer(obj, mode):
    return ("r" in mode and hasattr(obj, "read")) or ("w" in mode and hasattr(obj, "write"))


@contextmanager
def open_file(path_or_buf, mode="r"):
    if is_buffer(path_or_buf, mode):
        yield path_or_buf
    else:
        with open(path_or_buf, mode) as f:
            yield f


def get_fem_model_from_assembly(assembly: Assembly) -> Part:
    """
    Scans the assembly tree for parts containing FEM elements. If multiple FEM objects are not empty,
    they will be merged
    """
    parts = list(filter(lambda p: p.fem.is_empty() is False, assembly.get_all_parts_in_assembly(True)))

    if len(parts) > 1:
        raise ValueError(
            "This method does not yet support multipart FEM. Please make sure your assembly only contain 1 FEM"
        )
    elif len(parts) == 0:
        raise ValueError("At least 1 part must have a FEM mesh ")

    return parts[0]


def get_exe_path(fea_type: FEATypes):
    from ada.fem.formats.general import FEATypes

    if isinstance(fea_type, FEATypes):
        exe_name = fea_type.value
    else:
        exe_name = fea_type

    env_name = f"ADA_{exe_name}_exe"

    if Settings.fem_exe_paths.get(exe_name, None) is not None:
        exe_path = Settings.fem_exe_paths[exe_name]
    elif os.getenv(env_name):
        exe_path = os.getenv(env_name)
    elif shutil.which(f"{exe_name}"):
        exe_path = shutil.which(f"{exe_name}")
    elif shutil.which(f"{exe_name}.exe"):
        exe_path = shutil.which(f"{exe_name}.exe")
    elif shutil.which(f"{exe_name}.bat"):
        exe_path = shutil.which(f"{exe_name}.bat")
    else:
        raise FileNotFoundError(f'No Path to Executable "{exe_name}.exe" or "{exe_name}.bat" is found')

    exe_path = pathlib.Path(exe_path)

    if exe_path.exists() is False:
        return None

    return exe_path


def str_to_int(s):
    return int(float(s))


def str_to_float(s):
    from ada.core.utils import roundoff

    return roundoff(s)


def get_ff_regex(flag, *args):
    """
    Compile a regex search string for Fortran formatted string input.

    :param flag: Name of keyword flag (ie. the first word on a line of input parameters)
    :param args: Group name for each parameter. Include the character | to signify the parameters is optional.


    :return: Returns a compiled regex search string.. re.compile(..)
    """
    pattern_str = r"^(?P<flag>.*?)" if flag is True else rf"^{flag}"
    counter = 0
    re_in = re.IGNORECASE | re.MULTILINE | re.DOTALL

    def add_key(k):
        if "|" in k:
            return rf'(?: \s*(?P<{k.replace("|", "")}>.*?)|)'
        return rf" \s*(?P<{k}>.*?)"

    for i, key in enumerate(args):
        if type(key) is str:
            pattern_str += add_key(key)
            counter += 1
            if counter == 4 and i < len(args) - 1:
                counter = 0
                pattern_str += r"(?:\n|)\s*"
        elif type(key) is list:
            for subkey in key:
                pattern_str += add_key(subkey)
            pattern_str += r"(?:\n|)\s*"
        else:
            raise ValueError(f'Unrecognized input type "{type(key)}"')

    if not args:
        pattern_str += add_key("bulk")

    pattern_str += r"(?:(?=^[A-Z]|\Z))"

    return re.compile(pattern_str, re_in)


def _overwrite_dir(analysis_dir):
    print("Removing old files before copying new")
    try:
        if Settings.safe_deletion is True:
            send2trash(analysis_dir)
        else:
            shutil.rmtree(analysis_dir)
    except WindowsError as e:
        print(f"Failed to delete due to '{e}'")

    os.makedirs(analysis_dir, exist_ok=True)


def _lock_check(analysis_dir):
    lck_file = (analysis_dir / analysis_dir.stem).with_suffix(".lck")
    if lck_file.is_file():
        raise IOError(
            f'Found lock-file:\n\n"{lck_file}"\n\nThis indicates that an analysis is running.'
            "Please stop analysis and try again"
        )
    if (analysis_dir / "ada.lck").is_file():
        raise IOError(
            f'Found ADA lock-file:\n\n"{analysis_dir}\\ada.lck"\n\nThis indicates that the analysis folder is'
            f" locked Please removed ADA lock file if this is not the case and try again"
        )


def folder_prep(scratch_dir, analysis_name, overwrite):
    if scratch_dir is None:
        scratch_dir = pathlib.Path(Settings.scratch_dir)
    else:
        scratch_dir = pathlib.Path(scratch_dir)

    analysis_dir = scratch_dir / analysis_name
    if analysis_dir.is_dir():
        _lock_check(analysis_dir)
        if overwrite is True:
            _overwrite_dir(analysis_dir)
        else:
            raise IOError('The analysis folder exists. Please remove folder or pass argument "overwrite=True"')

    os.makedirs(analysis_dir, exist_ok=True)
    return analysis_dir


def run_windows(exe: LocalExecute, run_cmd, stop_cmd=None, exit_after=True, bat_start_str=None, run_in_shell=False):
    if bat_start_str is None:
        bat_start_str = f"""echo OFF
for %%* in (.) do set CurrDirName=%%~nx*
title %CurrDirName%
cd /d {exe.analysis_dir}
echo ON\ncall {run_cmd}"""

    if exit_after is False:
        bat_start_str += "\npause"

    start_bat = "run.bat"
    stop_bat = "stop.bat"

    os.makedirs(exe.execute_dir, exist_ok=True)

    with open(exe.execute_dir / start_bat, "w") as d:
        d.write(bat_start_str + "\nEXIT")

    if stop_cmd is not None:
        with open(exe.execute_dir / stop_bat, "w") as d:
            d.write(f"cd /d {exe.analysis_dir}\n{stop_cmd}")

    if Settings.execute_dir is not None:
        shutil.copy(exe.execute_dir / start_bat, Settings.execute_dir / start_bat)
        shutil.copy(exe.execute_dir / stop_bat, Settings.execute_dir / stop_bat)

    # If the script should be running from batch files, then this can be used
    if run_in_shell:
        _ = "start " + start_bat if exe.run_ext is True else "start /wait " + start_bat
    else:
        _ = "start " + start_bat if exe.run_ext is True else "call " + start_bat

    return run_tool(exe, run_cmd, "Windows")


def run_linux(exe, run_cmd):
    return run_tool(exe, run_cmd.split(), "Linux")


def run_tool(exe: LocalExecute, run_cmd, platform):
    fem_tool_name = type(exe).__name__.replace("Execute", "")
    props = dict(cwd=exe.execute_dir, env=os.environ, universal_newlines=True, encoding="utf-8")
    if exe.auto_execute is False:
        return None

    print(80 * "-")
    print(f'Starting {fem_tool_name} simulation "{exe.analysis_name}" (on {platform}) using {exe.cpus} cpus')
    if exe.run_ext is True:
        out = subprocess.Popen(run_cmd, **props)
        print(f"Note! This starts {fem_tool_name} in an external window on a separate thread.")
    else:
        # run_directly_on_windows(run_cmd, props, exe)
        props["capture_output"] = True
        out = subprocess.run(run_cmd, **props)
        print(f'Finished {fem_tool_name} simulation "{exe.analysis_name}"')
    print(80 * "-")
    return out


def run_directly_on_windows(run_cmd, props, exe):
    rstr = ""
    for out in execute(cmd=run_cmd, **props):
        print(out)
        rstr += out
    with open(exe.execute_dir / "run.log", "w") as f:
        f.write(rstr)


def execute(cmd, cwd, encoding, **kwargs):
    popen = subprocess.Popen(cmd, stdout=subprocess.PIPE, cwd=cwd, encoding=encoding)
    for stdout_line in popen.stdout:
        try:
            yield stdout_line.strip()
        except UnicodeDecodeError as e:
            logger.error(e)
            continue

    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, cmd)


def run_macOS(exe, run_cmd):
    raise NotImplementedError()


def interpret_fem_format_from_path(fem_path: str | pathlib.Path):
    from ada.fem.formats.general import FEATypes

    suffix = pathlib.Path(fem_path).suffix.lower()

    fem_type = None
    if suffix in (".fem", ".sif"):
        fem_type = FEATypes.SESAM
    elif suffix == ".inp":
        fem_type = FEATypes.ABAQUS
    elif suffix in (".frd",):
        fem_type = FEATypes.CALCULIX
    elif suffix in (".rmed", ".med"):
        fem_type = FEATypes.CODE_ASTER
    else:
        logger.error(f'unrecognized suffix "{suffix}"')

    return fem_type


def should_convert(res_path, overwrite):
    run_convert = True
    if res_path is not None:
        if res_path.exists() is True:
            run_convert = False
    if run_convert is True or overwrite is True:
        return True
    else:
        return False


def convert_shell_elem_to_plates(elem: Elem, parent: Part) -> list[Plate]:
    from ada import Plate
    from ada.core.vector_utils import is_coplanar

    plates = []
    fem_sec = elem.fem_sec
    fem_sec.material.parent = parent
    if len(elem.nodes) == 4:
        if is_coplanar(
            *elem.nodes[0].p,
            *elem.nodes[1].p,
            *elem.nodes[2].p,
            *elem.nodes[3].p,
        ):
            plates.append(
                Plate(
                    f"sh{elem.id}",
                    [n.p for n in elem.nodes],
                    fem_sec.thickness,
                    mat=fem_sec.material,
                    use3dnodes=True,
                    parent=parent,
                )
            )
        else:
            el_n1 = [elem.nodes[0].p, elem.nodes[1].p, elem.nodes[2].p]
            el_n2 = [elem.nodes[0].p, elem.nodes[2].p, elem.nodes[3].p]
            plates.append(
                Plate(f"sh{elem.id}", el_n1, fem_sec.thickness, mat=fem_sec.material, use3dnodes=True, parent=parent)
            )
            plates.append(
                Plate(
                    f"sh{elem.id}_1",
                    el_n2,
                    fem_sec.thickness,
                    use3dnodes=True,
                    mat=fem_sec.material,
                    parent=parent,
                )
            )
    else:
        plates.append(
            Plate(
                f"sh{elem.id}",
                [n.p for n in elem.nodes],
                fem_sec.thickness,
                mat=fem_sec.material,
                use3dnodes=True,
                parent=parent,
            )
        )
    return plates


def convert_part_shell_elements_to_plates(p: Part) -> Plates:
    return Plates(
        list(chain.from_iterable([convert_shell_elem_to_plates(sh, p) for sh in p.fem.elements.shell])), parent=p
    )


def convert_part_elem_bm_to_beams(p: Part) -> Beams:
    return Beams([line_elem_to_beam(bm, p) for bm in p.fem.elements.lines], parent=p)


def line_elem_to_beam(elem: Elem, parent: Part) -> Beam:
    """Convert FEM line element to Beam"""
    from ada import Beam

    a = parent.get_assembly()

    n1 = elem.nodes[0]
    n2 = elem.nodes[-1]
    e1 = None
    e2 = None
    elem.fem_sec.material.parent = parent
    if a.convert_options.fem2concepts_include_ecc is True:
        if elem.eccentricity is not None:
            ecc = elem.eccentricity
            if ecc.end1 is not None and ecc.end1.node.id == n1.id:
                e1 = ecc.end1.ecc_vector
            if ecc.end2 is not None and ecc.end2.node.id == n2.id:
                e2 = ecc.end2.ecc_vector

    if elem.fem_sec.section.type == "GENBEAM":
        logger.error(f"Beam elem {elem.id}  uses a GENBEAM which might not represent an actual cross section")

    return Beam(
        f"bm{elem.id}",
        n1,
        n2,
        sec=elem.fem_sec.section,
        mat=elem.fem_sec.material,
        up=elem.fem_sec.local_z,
        e1=e1,
        e2=e2,
        parent=parent,
    )


def convert_part_objects(p: Part, skip_plates, skip_beams):
    if skip_plates is False:
        p._plates = convert_part_shell_elements_to_plates(p)
    if skip_beams is False:
        p._beams = convert_part_elem_bm_to_beams(p)


def default_fem_res_path(
    name, scratch_dir=None, analysis_dir=None, fem_format=None
) -> dict[FEATypes, pathlib.Path] | str:
    from ada.fem.formats.general import FEATypes

    if scratch_dir is None and analysis_dir is None:
        scratch_dir = Settings.scratch_dir

    base_path = scratch_dir / name / name if analysis_dir is None else analysis_dir / name
    fem_format_map = {
        FEATypes.CODE_ASTER: base_path.with_suffix(".rmed"),
        FEATypes.ABAQUS: base_path.with_suffix(".pckle"),
        FEATypes.CALCULIX: base_path.with_suffix(".frd"),
        FEATypes.SESAM: (base_path.parent / f"{name}R1").with_suffix(".SIF"),
        FEATypes.USFOS: base_path.with_suffix(".fem"),
        FEATypes.XDMF: base_path.with_suffix(".xdmf"),
    }

    if fem_format is None:
        return fem_format_map

    return fem_format_map.get(fem_format)


def default_fem_inp_path(name, scratch_dir=None, analysis_dir=None):
    from ada.fem.formats.general import FEATypes

    base_path = scratch_dir / name / name if analysis_dir is None else analysis_dir / name
    return {
        FEATypes.CODE_ASTER: base_path.with_suffix(".export"),
        FEATypes.ABAQUS: base_path.with_suffix(".inp"),
        FEATypes.CALCULIX: base_path.with_suffix(".inp"),
        FEATypes.SESAM: (base_path.parent / f"{name}T1").with_suffix(".FEM"),
        FEATypes.USFOS: base_path.with_suffix(".raf"),
    }



from __future__ import annotations

from typing import TYPE_CHECKING

import ifcopenshell
import numpy as np

from ada.base.units import Units
from ada.config import get_logger
from ada.core.constants import O, X, Z
from ada.core.curve_utils import get_center_from_3_points_and_radius
from ada.core.vector_utils import (
    angle_between,
    calc_yvec,
    calc_zvec,
    global_2_local_nodes,
    unit_vector,
    vector_length,
)
from ada.ifc.utils import (
    create_guid,
    create_ifc_placement,
    create_ifcpolyline,
    create_local_placement,
    tesselate_shape,
    to_real,
    write_elem_property_sets,
)

if TYPE_CHECKING:
    from ada import Pipe, PipeSegElbow, PipeSegStraight

logger = get_logger()


def write_ifc_pipe(pipe: Pipe):
    ifc_pipe = write_pipe_ifc_elem(pipe)

    a = pipe.get_assembly()
    ifc_store = a.ifc_store
    f = ifc_store.f

    segments = []
    for param_seg in pipe.segments:
        res = write_pipe_segment(param_seg)
        if res is None:
            logger.error(f'Branch "{param_seg.name}" was not converted to ifc element')
        f.add(res)
        segments += [res]

    ifc_store.writer.add_related_elements_to_spatial_container(segments, ifc_pipe.GlobalId)

    return ifc_pipe


def write_pipe_segment(segment: PipeSegElbow | PipeSegStraight) -> ifcopenshell.entity_instance:
    from ada import PipeSegElbow, PipeSegStraight

    if isinstance(segment, PipeSegElbow):
        pipe_seg = write_pipe_elbow_seg(segment)
    elif isinstance(segment, PipeSegStraight):
        pipe_seg = write_pipe_straight_seg(segment)
    else:
        raise ValueError(f'Unrecognized Pipe Segment type "{type(segment)}"')

    assembly = segment.get_assembly()
    ifc_store = assembly.ifc_store
    ifc_store.writer.associate_elem_with_material(segment.material, pipe_seg)

    return pipe_seg


def write_pipe_ifc_elem(pipe: Pipe):
    if pipe.parent is None:
        raise ValueError("Cannot build ifc element without parent")

    a = pipe.get_assembly()
    f = a.ifc_store.f

    owner_history = a.ifc_store.owner_history
    parent = f.by_guid(pipe.parent.guid)

    placement = create_local_placement(
        f,
        origin=pipe.n1.p.astype(float).tolist(),
        loc_x=X,
        loc_z=Z,
        relative_to=parent.ObjectPlacement,
    )

    ifc_elem = f.create_entity(
        "IfcSpatialZone",
        pipe.guid,
        owner_history,
        pipe.name,
        "Description",
        None,
        placement,
        None,
        None,
        None,
    )

    f.createIfcRelAggregates(
        create_guid(),
        owner_history,
        "Site Container",
        None,
        parent,
        [ifc_elem],
    )

    return ifc_elem


def write_pipe_straight_seg(pipe_seg: PipeSegStraight):
    if pipe_seg.parent is None:
        raise ValueError("Parent cannot be None for IFC export")

    assembly = pipe_seg.parent.get_assembly()
    ifc_store = assembly.ifc_store
    f = ifc_store.f
    owner_history = ifc_store.owner_history

    p1 = pipe_seg.p1
    p2 = pipe_seg.p2

    ifcdir = f.createIfcDirection((0.0, 0.0, 1.0))

    rp1 = to_real(p1.p)
    rp2 = to_real(p2.p)
    xvec = unit_vector(p2.p - p1.p)
    a = angle_between(xvec, np.array([0, 0, 1]))
    zvec = np.array([0, 0, 1]) if a != np.pi and a != 0 else np.array([1, 0, 0])
    yvec = unit_vector(np.cross(zvec, xvec))
    seg_l = vector_length(p2.p - p1.p)

    extrusion_placement = create_ifc_placement(f, (0.0, 0.0, 0.0), (0.0, 0.0, 1.0), (1.0, 0.0, 0.0))

    section_profile = ifc_store.get_profile_def(pipe_seg.section)
    if section_profile is None:
        raise ValueError("Section profile not found")

    solid = f.createIfcExtrudedAreaSolid(section_profile, extrusion_placement, ifcdir, seg_l)

    polyline = create_ifcpolyline(f, [rp1, rp2])

    axis_representation = f.createIfcShapeRepresentation(ifc_store.get_context("Axis"), "Axis", "Curve3D", [polyline])
    body_representation = f.createIfcShapeRepresentation(ifc_store.get_context("Body"), "Body", "SweptSolid", [solid])

    product_shape = f.createIfcProductDefinitionShape(None, None, [axis_representation, body_representation])

    origin = f.createIfcCartesianPoint(O)
    local_z = f.createIfcDirection(Z)
    local_x = f.createIfcDirection(X)
    d237 = f.createIfcLocalPlacement(None, f.createIfcAxis2Placement3D(origin, local_z, local_x))

    d256 = f.createIfcCartesianPoint(rp1)
    d257 = f.createIfcDirection(to_real(xvec))
    d258 = f.createIfcDirection(to_real(yvec))
    d236 = f.createIfcAxis2Placement3D(d256, d257, d258)
    local_placement = f.createIfcLocalPlacement(d237, d236)

    pipe_segment = f.createIfcPipeSegment(
        pipe_seg.guid,
        owner_history,
        pipe_seg.name,
        "An awesome pipe",
        None,
        local_placement,
        product_shape,
        None,
    )

    return pipe_segment


def write_pipe_elbow_seg(pipe_elbow: PipeSegElbow):
    if pipe_elbow.parent is None:
        raise ValueError("Parent cannot be None for IFC export")

    a = pipe_elbow.parent.get_assembly()
    f = a.ifc_store.f

    owner_history = a.ifc_store.owner_history

    tol = Units.get_general_point_tol(a.units)

    ifc_elbow = elbow_revolved_solid(pipe_elbow, f, tol)

    pfitting_placement = create_local_placement(f)

    pfitting = f.create_entity(
        "IfcPipeFitting",
        pipe_elbow.guid,
        owner_history,
        pipe_elbow.name,
        "An curved pipe segment",
        None,
        pfitting_placement,
        ifc_elbow,
        None,
        None,
    )

    props = dict(
        bend_radius=pipe_elbow.bend_radius,
        p1=pipe_elbow.arc_seg.p1,
        p2=pipe_elbow.arc_seg.p2,
        midpoint=pipe_elbow.arc_seg.midpoint,
    )

    write_elem_property_sets(props, pfitting, f, owner_history)

    return pfitting


def elbow_tesselated(self: PipeSegElbow, f, schema, a):
    shape = self.solid

    if shape is None:
        logger.error(f"Unable to create geometry for Branch {self.name}")
        return None
    point_tol = Units.get_general_point_tol(a.units)
    serialized_geom = tesselate_shape(shape, schema, point_tol)
    ifc_shape = f.add(serialized_geom)

    return ifc_shape


def elbow_revolved_solid(elbow: PipeSegElbow, f, tol=1e-1):
    xvec1 = unit_vector(elbow.xvec1)
    xvec2 = unit_vector(elbow.xvec2)
    normal = unit_vector(calc_zvec(xvec1, xvec2))
    p1, p2, p3 = elbow.p1.p, elbow.p2.p, elbow.p3.p

    a = elbow.get_assembly()
    ifc_store = a.ifc_store

    # Profile
    profile = ifc_store.get_profile_def(elbow.section)

    # Revolve Angle
    revolve_angle = np.rad2deg(angle_between(xvec1, xvec2))

    # Revolve Point
    cd = get_center_from_3_points_and_radius(p1, p2, p3, elbow.bend_radius, tol=tol)
    diff = cd.center - elbow.arc_seg.p1

    # Transform Axis normal and position to the local coordinate system
    yvec = calc_yvec(normal, xvec1)
    new_csys = (normal, yvec, xvec1)

    diff_tra = global_2_local_nodes(new_csys, O, [diff])[0]
    n_tra = global_2_local_nodes(new_csys, O, [normal])[0]

    n_tra_norm = to_real(unit_vector(n_tra))
    diff_tra_norm = to_real(diff_tra)

    # Revolve Axis
    rev_axis_dir = f.create_entity("IfcDirection", n_tra_norm)
    revolve_point = f.create_entity("IfcCartesianPoint", diff_tra_norm)
    revolve_axis1 = f.create_entity("IfcAxis1Placement", revolve_point, rev_axis_dir)

    position = create_ifc_placement(f, elbow.arc_seg.p1, xvec1, normal)

    # Body representation

    ifc_shape = f.create_entity("IfcRevolvedAreaSolid", profile, position, revolve_axis1, revolve_angle)
    body = f.create_entity("IfcShapeRepresentation", ifc_store.get_context("Body"), "Body", "SweptSolid", [ifc_shape])

    # Axis representation
    polyline = create_ifcpolyline(f, [p1, p2, p3])
    axis = f.create_entity("IfcShapeRepresentation", ifc_store.get_context("Axis"), "Axis", "Curve3D", [polyline])

    # Final Product Shape
    prod_def_shp = f.create_entity("IfcProductDefinitionShape", None, None, (body, axis))

    return prod_def_shp



from __future__ import annotations

import re
from itertools import chain
from typing import TYPE_CHECKING, Iterable

from ada.config import get_logger
from ada.core.utils import Counter, roundoff
from ada.fem import ConnectorSection, FemSection
from ada.fem.containers import FemSections
from ada.fem.elements import Eccentricity
from ada.fem.shapes import ElemType

from . import cards
from .helper_utils import list_cleanup

part_name_counter = Counter(1, "Part")
_re_in = re.IGNORECASE | re.MULTILINE | re.DOTALL

if TYPE_CHECKING:
    from ada.concepts.spatial import Assembly
    from ada.fem import FEM

logger = get_logger()


def get_sections_from_inp(bulk_str, fem: FEM) -> FemSections:
    iter_beams = get_beam_sections_from_inp(bulk_str, fem)
    iter_shell = get_shell_sections_from_inp(bulk_str, fem)
    iter_solid = get_solid_sections_from_inp(bulk_str, fem)

    return FemSections(chain.from_iterable([iter_beams, iter_shell, iter_solid]), fem)


def get_beam_sections_from_inp(bulk_str: str, fem: FEM) -> Iterable[FemSection]:
    # Source:  https://abaqus-docs.mit.edu/2017/English/SIMACAEELMRefMap/simaelm-c-beamcrosssectlib.htm
    from ada import Section
    from ada.sections import GeneralProperties

    ass = fem.parent.get_assembly()
    if bulk_str.lower().find("*beam section") == -1:
        return []

    def interpret_section(profile_name, sec_type, props):
        props_clean = [roundoff(x) for x in filter(lambda x: x.strip() != "", props.split(","))]
        if sec_type.upper() == "BOX":
            b, h, t1, t2, t3, t4 = props_clean
            return Section(
                profile_name,
                "BG",
                h=h,
                w_btn=b,
                w_top=b,
                t_w=t1,
                t_fbtn=t4,
                t_ftop=t2,
                parent=fem,
            )
        elif sec_type.upper() == "CIRC":
            return Section(profile_name, "CIRC", r=props_clean[0], parent=fem)
        elif sec_type.upper() == "I":
            (
                l,
                h,
                b1,
                b2,
                t1,
                t2,
                t3,
            ) = props_clean
            return Section(
                profile_name,
                "IG",
                h=h,
                w_btn=b1,
                w_top=b2,
                t_w=t3,
                t_fbtn=t1,
                t_ftop=t2,
                parent=fem,
            )
        elif sec_type.upper() == "L":
            b, h, t1, t2 = props_clean
            return Section(profile_name, "HP", h=h, w_btn=b, t_w=t2, t_fbtn=t1, parent=fem)
        elif sec_type.upper() == "PIPE":
            r, t = props_clean
            return Section(profile_name, "TUB", r=r, wt=t, parent=fem)
        elif sec_type.upper() == "TRAPEZOID":
            # Currently converts Trapezoid to general beam
            b, h, a, d = props_clean
            # Assuming the Abaqus trapezoid element is symmetrical
            c = (b - a) / 2

            # The properties were quickly copied from a resource online. Most likely it contains error
            # https: // www.efunda.com / math / areas / trapezoidJz.cfm
            genprops = GeneralProperties(
                Ax=h * (a + b) / 2,
                Ix=h
                * (
                    b * h**2
                    + 3 * a * h**2
                    + a**3
                    + 3 * a * c**2
                    + 3 * c * a**2
                    + b**3
                    + c * b**2
                    + a * b**2
                    + b * c**2
                    + 2 * a * b * c
                    + b * a**2
                ),
                Iy=(h**3) * (3 * a + b) / 12,
                Iz=h
                * (
                    a**3
                    + 3 * a * c**2
                    + 3 * c * a**2
                    + b**3
                    + c * b**2
                    + a * b**2
                    + 2 * a * b * c
                    + b * a**2
                )
                / 12,
            )
            return Section(profile_name, "GENBEAM", genprops=genprops, parent=fem)
        else:
            logger.error(f'Currently unsupported section type "{sec_type}". Will return None')
            return None

    def grab_beam(match):
        d = match.groupdict()
        elset = fem.elsets[d["elset"]]
        name = elset.name
        profile_name = elset.name
        material = ass.materials.get_by_name(d["material"])
        # material = parent.parent.materials.get_by_name(d['material'])
        temperature = d["temperature"]
        section_type = d["sec_type"]
        geo_props = d["line1"]
        sec = interpret_section(profile_name, section_type, geo_props)
        if sec is None:
            return None
        beam_y = [float(x.strip()) for x in d["line2"].split(",") if x.strip() != ""]
        metadata = dict(
            temperature=temperature,
            profile=profile_name.strip(),
            section_type=section_type,
            line1=geo_props,
        )
        res = fem.parent.sections.add(sec)
        if res is not None:
            sec = res
        return FemSection(
            name.strip(),
            sec_type=ElemType.LINE,
            elset=elset,
            section=sec,
            local_y=beam_y,
            material=material,
            metadata=metadata,
            parent=fem,
        )

    return filter(lambda x: x is not None, map(grab_beam, cards.re_beam.finditer(bulk_str)))


def get_solid_sections_from_inp(bulk_str, fem: FEM):
    secnames = Counter(1, "solidsec")
    a = fem.parent.get_assembly()
    if bulk_str.lower().find("*solid section") == -1:
        return []

    solid_iter = cards.re_solid.finditer(bulk_str)

    def grab_solid(m_in):
        name = m_in.group(1) if m_in.group(1) is not None else next(secnames)
        elset = m_in.group(2)
        material = m_in.group(3)
        mat = a.materials.get_by_name(material)
        return FemSection(
            name=name,
            sec_type=ElemType.SOLID,
            elset=elset,
            material=mat,
            parent=fem,
        )

    return map(grab_solid, solid_iter)


def get_shell_sections_from_inp(bulk_str, fem: FEM) -> Iterable[FemSection]:
    if bulk_str.lower().find("*shell section") == -1:
        return []

    a = fem.parent.get_assembly()
    sh_name = Counter(1, "sh")
    return (get_shell_section(m, sh_name, fem, a) for m in cards.re_shell.finditer(bulk_str))


def get_shell_section(m, sh_name, fem: "FEM", a: "Assembly"):
    d = m.groupdict()
    name = next(sh_name)
    elset = fem.sets.get_elset_from_name(d["elset"])

    material = d["material"]
    mat = a.materials.get_by_name(material)
    thickness = float(d["t"])
    offset = d["offset"]
    if offset is not None:
        # TODO: update this with the latest eccentricity class
        logger.warning("Offset for Shell elements is not yet evaluated")
        for el in elset.members:
            el.eccentricity = Eccentricity(sh_ecc_vector=offset)
    int_points = d["int_points"]
    metadata = dict(controls=d["controls"])

    return FemSection(
        name=name,
        sec_type=ElemType.SHELL,
        thickness=thickness,
        elset=elset,
        material=mat,
        int_points=int_points,
        parent=fem,
        metadata=metadata,
    )


def get_connector_sections_from_bulk(bulk_str: str, parent: FEM) -> dict[str, ConnectorSection]:
    import numpy as np

    consecsd = dict()

    for m in cards.connector_behaviour.regex.finditer(bulk_str):
        d = m.groupdict()
        name = d["name"]
        comp = int(d["component"])

        res = np.fromstring(list_cleanup(d["bulk"]), sep=",", dtype=np.float64)
        size = res.size
        cols = comp + 1
        rows = int(size / cols)
        res_ = res.reshape(rows, cols)
        consecsd[name] = ConnectorSection(name, [res_], [], metadata=d, parent=parent)
    return consecsd



from ..utils import create_guid


def write_door(f, owner_history, insert_placement, insert_shape):
    return f.createIfcDoor(
        create_guid(),
        owner_history,
        "Door",
        "An awesome Door",
        None,
        insert_placement,
        insert_shape,
        None,
        None,
    )


def write_window(f, owner_history, insert_placement, insert_shape):
    return f.create_entity(
        "IfcWindow",
        create_guid(),
        owner_history,
        "Window",
        "An awesome window",
        None,
        insert_placement,
        insert_shape,
        None,
        None,
    )



from typing import TYPE_CHECKING

import numpy as np

from ada.visualize.renderer_occ import occ_shape_to_faces

if TYPE_CHECKING:
    from ada import Assembly


def to_gltf(assembly: "Assembly", output_file_path):
    quality = 1.0
    render_edges = False
    parallel = True

    for p in assembly.parts.values():
        vertices = None
        normals = None
        indices = None
        id_map = dict()
        for obj in p.get_all_physical_objects():
            geom = obj.solid()
            np_vertices, np_faces, np_normals, _ = occ_shape_to_faces(geom, quality, render_edges, parallel)
            if vertices is None:
                vertices = np_vertices
                normals = np_normals
                indices = np_faces
                id_map[obj.guid] = (int(indices[0]), int(indices[-1]))
            else:
                vertices = np.concatenate([vertices, np_vertices])
                normals = np.concatenate([normals, np_normals])
                adjusted_indices = np_faces + len(indices)
                indices = np.concatenate([indices, adjusted_indices])
                id_map[obj.guid] = (int(adjusted_indices[0]), int(adjusted_indices[-1]))

    raise NotImplementedError("Export to GLTF is not yet supported")



from typing import TYPE_CHECKING

import numpy as np

from ada.config import Settings

from .write_sets import _add_node_sets

if TYPE_CHECKING:
    from ada import Part


def _write_nodes(part: "Part", time_step, profile, families):
    """

    TODO: Go through each data group and set in HDF5 file and make sure that it writes what was read 1:1.
        Use cylinder.med as a benchmark.

    Add the following datasets ['COO', 'FAM', 'NUM'] to the 'NOE' group

    :param part:
    :param time_step:
    :param profile:
    :return:
    """
    points = np.zeros((int(part.fem.nodes.max_nid), 3))

    def pmap(n):
        points[int(n.id - 1)] = n.p

    list(map(pmap, part.fem.nodes))

    # Try this
    if Settings.ca_experimental_id_numbering is True:
        points = np.array([n.p for n in part.fem.nodes])

    nodes_group = time_step.create_group("NOE")
    nodes_group.attrs.create("CGT", 1)
    nodes_group.attrs.create("CGS", 1)

    nodes_group.attrs.create("PFL", np.string_(profile))
    coo = nodes_group.create_dataset("COO", data=points.flatten(order="F"))
    coo.attrs.create("CGT", 1)
    coo.attrs.create("NBR", len(points))

    if Settings.ca_experimental_id_numbering is True:
        node_ids = [n.id for n in part.fem.nodes]
        num = nodes_group.create_dataset("NUM", data=node_ids)
        num.attrs.create("CGT", 1)
        num.attrs.create("NBR", len(points))

    if len(part.fem.nsets.keys()) > 0:
        _add_node_sets(nodes_group, part, points, families)

    assembly = part.get_assembly()

    if assembly != part and len(assembly.fem.nsets.keys()) > 0:
        _add_node_sets(nodes_group, assembly, points, families)



from typing import TYPE_CHECKING, List

from ada.fem import Load, LoadGravity

if TYPE_CHECKING:
    from ada import FEM


def load_str(load: Load):
    if isinstance(load, LoadGravity):
        return write_gravity_load_str(load)
    else:
        raise ValueError("Calculix does not accept Loads without reference to a fem_set")


def write_gravity_load_str(load: LoadGravity):
    dof = [0, 0, 1] if load.dof is None else load.dof
    fem_set = load.fem_set.name
    return f"""** Name: gravity   Type: Gravity
*Dload
{fem_set}, GRAV, {load.magnitude}, {', '.join([str(x) for x in dof[:3]])}"""


def check_if_grav_loads(fem: "FEM"):
    if LoadGravity in [type(step) for step in fem.steps]:
        return True
    else:
        return False


def get_all_grav_loads(fem: "FEM") -> List[LoadGravity]:
    return list(filter(lambda x: isinstance(x, LoadGravity), fem.get_all_loads()))



from __future__ import annotations

import os
import pathlib
import subprocess

from ada.config import get_logger
from ada.core.file_system import get_short_path_name

logger = get_logger()


def convert_sin_to_sif(sin_file: str | pathlib.Path, use_siu=False) -> None:
    if isinstance(sin_file, str):
        sin_file = pathlib.Path(sin_file)

    prepost_exe = os.environ.get("ADA_prepost_exe", None)
    if prepost_exe is None:
        raise FileNotFoundError("Prepost executable is not set. Please set it using `ADA_prepost_exe`")

    analysis_name = sin_file.stem
    formatting = "SIF-FORMATTED" if use_siu is False else "SIU-UNFORMATTED"
    jnl_file_str = f"OPEN SIN-DIRECT-ACCESS '' {analysis_name} OLD READ-ONLY\n"
    jnl_file_str += f"WRITE {formatting} '' {analysis_name} 1\nEND\nEXIT"

    with open(sin_file.parent / "run_prepost.jnl", "w") as f:
        f.write(jnl_file_str)

    run_params = "/NAME=PREPOST1/STAT=NEW/FORCED/LICENSE-WAIT=ON"
    log_params = f"more < {analysis_name}.JNL > log_{analysis_name}.log & echo finished prepost run 1"

    exe_str = f"set EXEPATH={get_short_path_name(str(prepost_exe))}"

    run_str = f"{exe_str}\nstart /w %EXEPATH% /INTER=L/COM-FI=run_prepost.jnl {run_params} & {log_params}"

    run_bat_file = sin_file.parent / "run_sin2sif.bat"
    with open(run_bat_file, "w") as f:
        f.write(run_str)

    props = dict(shell=True, cwd=sin_file.parent, env=os.environ, universal_newlines=True)
    props["capture_output"] = True
    out = subprocess.run(str(run_bat_file), **props)
    logger.info(f'Finished SIN2SIF operation on "{analysis_name}"')

    res_str = str(out.stderr + out.stdout)
    with open(sin_file.parent / "run_prepost_log.txt", "w") as f:
        f.write(res_str)

    return out



from __future__ import annotations

from typing import TYPE_CHECKING

from ada.fem import StepEigen, StepImplicit

if TYPE_CHECKING:
    from ada.concepts.spatial import Part

from .steps import eigen, static


def create_step_str(step: StepEigen | StepImplicit, part: Part) -> str:
    st = StepEigen.TYPES
    step_map = {st.STATIC: static.step_static_str, st.EIGEN: eigen.step_eig_str}

    step_writer = step_map.get(step.type, None)

    if step_writer is None:
        raise NotImplementedError(f'Step type "{step.type}" is not yet supported')

    return step_writer(step, part)



# An example of Eurocode equations
from __future__ import annotations

import math
from typing import TYPE_CHECKING

from ada.sections.categories import BaseTypes

if TYPE_CHECKING:
    from ada.fem import Elem


def ec3_654(elem: Elem, forces: list[float], buckling_length) -> float:
    if len(forces) != 6:
        raise ValueError("Length of Forces must be 6")

    m_ed = forces[4]
    p = elem.fem_sec.section.properties

    wy = p.Wymin
    fy = elem.fem_sec.material.model.sig_y
    x_lt = lat_buckling_xlt(elem, buckling_length)

    gam_m1 = 1.15
    mb_rd = x_lt * wy * fy / gam_m1
    return m_ed / mb_rd


def lat_buckling_xlt(elem, buckling_length, is_welded=False):
    s = elem.fem_sec.section
    p = s.properties
    wy = p.Wymin
    fy = elem.fem_sec.material.model.sig_y

    imp_fac = dict(a=0.21, b=0.34, c=0.49, d=0.76)
    if elem.fem_sec.section.type == BaseTypes.IPROFILE:
        if is_welded:
            if s.h / s.w_btn <= 2:
                a_lt = imp_fac["a"]
            else:
                a_lt = imp_fac["b"]
        else:
            if s.h / s.w_btn <= 2:
                a_lt = imp_fac["c"]
            else:
                a_lt = imp_fac["d"]
    else:
        a_lt = imp_fac["d"]

    m_cr = critical_moment(elem, buckling_length)
    lam_lt = math.sqrt(wy * fy / m_cr)
    phi_lt = 0.5 * (1 + a_lt * (lam_lt - 0.2) + lam_lt**2)
    xi_lt = 1.0 / (phi_lt + math.sqrt(phi_lt**2 - lam_lt**2))

    return min(xi_lt, 1.0)


def critical_moment(elem, length):
    # Todo: this is not at all correct. Will finish this later
    p = elem.fem_sec.section.properties

    g = elem.fem_sec.material.model.G
    e = elem.fem_sec.material.model.E
    if elem.fem_sec.section.type == BaseTypes.IPROFILE:
        m0_cr = (math.pi / length) * math.sqrt(g * p.Ix * p.Iz)
        i_w = 0
    else:
        raise NotImplementedError()

    my_cr = math.sqrt(1 + ((math.pi / length) ** 2) * (e * i_w / (g * p.Ix)))

    return m0_cr * my_cr



from __future__ import annotations

import xml.etree.ElementTree as ET
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ada import Material, Part


def add_materials(root: ET.Element, part: Part):
    materials_elem = ET.Element("materials")

    # Add the new element underneath <properties>
    root.append(materials_elem)

    for material in part.materials:
        add_isotropic_material(material, materials_elem)


def add_isotropic_material(material: Material, xml_root: ET.Element):
    section_elem = ET.Element("material", {"name": material.name})
    section_props = ET.Element(
        "isotropic_linear_material",
        dict(
            yield_stress=str(material.model.sig_y),
            density=str(material.model.rho),
            youngs_modulus=str(material.model.E),
            poissons_ratio=str(material.model.v),
            thermal_expansion=str(material.model.alpha),
            damping=str(material.model.zeta),
        ),
    )

    section_elem.append(section_props)
    xml_root.append(section_elem)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from ada.fem import Elem

from .shapes import ElemShape
from .shapes import definitions as shape_def

if TYPE_CHECKING:
    from ada import FEM, Assembly, Beam, Node, Part, Plate
    from ada.fem import FemSet


def get_eldata(fem_source: Assembly | Part | FEM):
    """Return a dictionary of basic mesh statistics"""
    from ada import FEM, Assembly, Part

    el_types = dict()

    def scan_elem(mesh):
        for el in mesh.elements:
            if el.type not in el_types.keys():
                el_types[el.type] = 1
            else:
                el_types[el.type] += 1

    if isinstance(fem_source, Assembly):
        for p in fem_source.parts.values():
            scan_elem(p.fem)
    elif issubclass(type(fem_source), Part):
        scan_elem(fem_source.fem)
    elif isinstance(fem_source, FEM):
        scan_elem(fem_source)
    else:
        raise ValueError(f'Unknown fem_source "{fem_source}"')
    return el_types


def get_beam_end_nodes(bm: Beam, end=1, tol=1e-3) -> list[Node]:
    """Get list of nodes from end of beam"""
    p = bm.parent
    nodes = p.fem.nodes
    w = bm.section.w_btn
    h = bm.section.h
    xv = np.array(bm.xvec)
    yv = np.array(bm.yvec)
    zv = np.array(bm.up)
    if end == 1:
        p = bm.n1.p
    else:
        p = bm.n2.p
    n_min = p - xv * tol - (h / 2 + tol) * zv - (w / 2 + tol) * yv
    n_max = p + xv * tol + (h / 2 + tol) * zv + (w / 2 + tol) * yv
    members = [e for e in nodes.get_by_volume(n_min, n_max)]
    return members


def get_nodes_along_plate_edges(pl: Plate, fem: FEM, edge_indices=None, tol=1e-3) -> list[Node]:
    """Return FEM nodes from edges of a plate"""

    res = []
    bmin, bmax = list(zip(*pl.bbox()))
    bmin_smaller = np.array(bmin) + pl.poly.xdir * tol + pl.poly.ydir * tol
    bmax_smaller = np.array(bmax) - pl.poly.xdir * tol - pl.poly.ydir * tol
    all_res = fem.nodes.get_by_volume(bmin, bmax)
    res = [n.id for n in fem.nodes.get_by_volume(bmin_smaller, bmax_smaller)]
    return list(filter(lambda x: x.id not in res, all_res))


def is_line_elem(elem: Elem):
    return True if isinstance(elem.type, shape_def.LineShapes) else False


def is_tri6_shell_elem(sh_fs):
    elem_check = [x.type == ElemShape.TYPES.shell.TRI6 for x in sh_fs.elset.members]
    return all(elem_check)


def is_quad8_shell_elem(sh_fs):
    elem_check = [x.type == ElemShape.TYPES.shell.QUAD8 for x in sh_fs.elset.members]
    return all(elem_check)


def is_parent_of_node_solid(no: Node) -> bool:
    refs = no.refs
    for elem in refs:
        if isinstance(elem.type, shape_def.SolidShapes):
            return True
    return False


def elset_to_part(name: str, elset: FemSet) -> Part:
    """Create a new part based on a specific element set."""
    from ada import Part

    fem = elset.parent
    p = Part(name)

    for mem in elset.members:
        if elset.type == elset.TYPES.ELSET:
            p.fem.add_elem(mem)
            if mem.fem_sec.name not in p.fem.sections.name_map.keys():
                p.fem.add_section(mem.fem_sec)
                fem.sections.remove(mem.fem_sec)
            if mem.fem_sec.elset.name not in p.fem.elsets.keys():
                p.fem.add_set(mem.fem_sec.elset)
                fem.sets.remove(mem.fem_sec.elset)
            for n in mem.nodes:
                if n not in p.fem.nodes:
                    p.fem.nodes.add(n)
            # fem.nodes.remove(mem.nodes)
        else:
            p.fem.nodes.add(mem)

    if elset.type == elset.TYPES.ELSET:
        fem.elements.remove_elements_by_set(elset)
    else:
        fem.nodes.remove(elset.members)

    return p


def split_line_element_in_two(el: Elem) -> Elem:
    from ada import Node

    n1 = el.nodes[0]
    n2 = el.nodes[-1]
    midp = (n1.p + n2.p) / 2
    new_node = el.parent.nodes.add(Node(midp))
    el.nodes[-1] = new_node
    elset = el.elset
    elem = Elem(None, [new_node, n2], el.type, elset=elset, fem_sec=el.fem_sec, parent=el.parent)
    fs = el.fem_sec
    fs.elset.add_members([elem])
    return elem



from __future__ import annotations

import os
import pathlib
from typing import TYPE_CHECKING

from ada import fem
from ada.base.units import Units
from ada.concepts.curves import ArcSegment, CurvePoly, CurveRevolve, LineSegment
from ada.concepts.fasteners import Bolts, Weld
from ada.concepts.piping import Pipe, PipeSegElbow, PipeSegStraight
from ada.concepts.points import Node
from ada.concepts.primitives import (
    Penetration,
    PrimBox,
    PrimCyl,
    PrimExtrude,
    PrimRevolve,
    PrimSphere,
    PrimSweep,
    Shape,
)
from ada.concepts.spatial import Assembly, Group, Part
from ada.concepts.stru_beams import Beam
from ada.concepts.stru_plates import Plate
from ada.concepts.stru_walls import Wall
from ada.concepts.transforms import Instance, Placement, Transform
from ada.concepts.user import User
from ada.fem import FEM
from ada.materials import Material
from ada.sections import Section

if TYPE_CHECKING:
    import ifcopenshell

    from ada.fem.formats.sesam.results.read_cc import CCData
    from ada.fem.results.common import FEAResult

__author__ = "Kristoffer H. Andersen"


def from_ifc(ifc_file: os.PathLike | ifcopenshell.file, units=Units.M, name="Ada") -> Assembly:
    if isinstance(ifc_file, (os.PathLike, str)):
        ifc_file = pathlib.Path(ifc_file).resolve().absolute()
        print(f'Reading "{ifc_file.name}"')
    else:
        print("Reading IFC file object")

    a = Assembly(units=units, name=name)
    a.read_ifc(ifc_file)
    return a


def from_step(step_file: str | pathlib.Path, source_units=Units.M, **kwargs) -> Assembly:
    a = Assembly()
    a.read_step_file(step_file, source_units=source_units, **kwargs)
    return a


def from_fem(
    fem_file: str | list | pathlib.Path,
    fem_format: str | list = None,
    name: str | list = None,
    enable_cache=False,
    source_units=Units.M,
    fem_converter="default",
) -> Assembly:
    a = Assembly(enable_cache=enable_cache, units=source_units)
    if type(fem_file) is str or issubclass(type(fem_file), pathlib.Path):
        a.read_fem(fem_file, fem_format, name, fem_converter=fem_converter)
    elif type(fem_file) is list:
        for i, f in enumerate(fem_file):
            fem_format_in = fem_format if fem_format is None else fem_format[i]
            name_in = name if name is None else name[i]
            a.read_fem(f, fem_format_in, name_in, fem_converter=fem_converter)
    else:
        raise ValueError(f'fem_file must be either string or list. Passed type was "{type(fem_file)}"')

    return a


def from_fem_res(fem_file: str | pathlib.Path, fem_format: str = None) -> FEAResult:
    from ada.fem.formats.postprocess import postprocess

    return postprocess(fem_file, fem_format)


def from_sesam_cc(fem_file: str | pathlib.Path) -> dict[str, CCData]:
    from ada.fem.formats.sesam.results.read_cc import read_cc_file

    return read_cc_file(fem_file)


def from_genie_xml(xml_path, **kwargs) -> Assembly:
    from ada.fem.formats.sesam.xml.store import GxmlStore

    gxml = GxmlStore(xml_path)
    p = gxml.to_part(**kwargs)
    return Assembly(name=kwargs.get("name", p.name)) / p


__all__ = [
    "Assembly",
    "Part",
    "FEM",
    "from_ifc",
    "from_fem",
    "Beam",
    "Group",
    "Plate",
    "Pipe",
    "PipeSegStraight",
    "PipeSegElbow",
    "Wall",
    "Penetration",
    "Section",
    "Material",
    "Shape",
    "Node",
    "Placement",
    "PrimBox",
    "PrimCyl",
    "PrimExtrude",
    "PrimRevolve",
    "PrimSphere",
    "PrimSweep",
    "CurvePoly",
    "CurveRevolve",
    "LineSegment",
    "ArcSegment",
    "Transform",
    "Instance",
    "User",
    "Bolts",
    "Weld",
    "fem",
]



from operator import attrgetter

from ada import FEM, Part
from ada.config import get_logger
from ada.core.utils import Counter
from ada.fem import Elem

logger = get_logger()


def shell_str(part: Part):
    pl_str = "'            Elem ID      np1      np2      np3      np4    mater   geom      ec1    ec2    ec3    ec4\n"
    sec_str = """'            Geom ID     Thick"""
    thick = []
    for fs in sorted(part.fem.sections.shells, key=attrgetter("id")):
        t = fs.thickness
        if t is None:
            raise ValueError("Thickness cannot be None")
        if fs not in thick:
            thick.append(fs)
            sec_str += "\n PLTHICK{:>12}{:>10}".format(fs.id, t)

    sec_str += "\n"

    def write_elem(el: Elem):
        if len(el.nodes) > 4:
            raise ValueError(f'Shell id "{el.id}" consist of {len(el.nodes)} nodes')
        else:
            nodes_str = "".join(["{:>9}".format(no.id) for no in el.nodes])
            if len(el.nodes) == 3:
                return f" TRISHELL{el.id:>11}{nodes_str}{'':>9}{el.fem_sec.material.id:>9}{el.fem_sec.id:>7}"
            else:
                return f" QUADSHEL{el.id:>11}{nodes_str}{el.fem_sec.material.id:>9}{el.fem_sec.id:>7}"

    return sec_str + pl_str + "\n".join(list(map(write_elem, sorted(part.fem.elements.shell, key=attrgetter("id")))))


def beam_str(fem: FEM, eccen):
    """

    # USFOS Strings

    # Beam String
    '            Elem ID     np1      np2   material   geom    lcoor    ecc1    ecc2
    BEAM            1127     1343     1344        1        1       1

    # Unit Vector String
    '            Loc-Coo           dx             dy             dz
    UNITVEC    60000001        0.00000        0.00000        1.00000

    """
    locvecs = []
    eccen_counter = Counter(1)
    loc_str = "'\n'            Loc-Coo           dx             dy             dz\n"
    bm_str = "'\n'            Elem ID     np1      np2   material   geom    lcoor    ecc1    ecc2\n"

    logger.info(
        "Note! Second order formulations of beam elements is not supported by Usfos beam. "
        "Will use regular beam formulation"
    )

    def write_elem(el: Elem) -> str:
        nonlocal locvecs
        n1 = el.nodes[0]
        n2 = el.nodes[1]
        fem_sec = el.fem_sec
        mat = fem_sec.material
        xvec = fem_sec.local_z
        xvec_str = f"{xvec[0]:>13.5f}{xvec[1]:>15.5f}{xvec[2]:>15.5f}"

        if xvec_str in locvecs:
            locid = locvecs.index(xvec_str)
        else:
            locvecs.append(xvec_str)
            locid = locvecs.index(xvec_str)

        if el.eccentricity is not None:
            ecc1_str = " 0"
            ecc2_str = " 0"
            if el.eccentricity.end1 is not None:
                ecc1 = next(eccen_counter)
                eccen.append((ecc1, el.eccentricity.end1.ecc_vector))
                ecc1_str = f" {ecc1}"
            if el.eccentricity.end2 is not None:
                ecc2 = next(eccen_counter)
                eccen.append((ecc2, el.eccentricity.end2.ecc_vector))
                ecc2_str = f" {ecc2}"
        else:
            ecc1_str = ""
            ecc2_str = ""
        return f" BEAM{el.id:>15}{n1.id:>8}{n2.id:>9}{mat.id:>11}{el.fem_sec.id:>7}{locid + 1:>9}{ecc1_str}{ecc2_str}"

    bm_str += "\n".join(list(map(write_elem, fem.elements.lines)))

    for i, loc in enumerate(locvecs):
        loc_str += " UNITVEC{:>13}{:<10}\n".format(i + 1, loc)

    return bm_str + "\n" + loc_str



from __future__ import annotations

import pathlib
from dataclasses import dataclass
from itertools import groupby

import h5py
import numpy as np


@dataclass
class CCData:
    member_name: str
    code_check_name: str
    components: list[str]
    values: list[np.ndarray]

    def get_component_data(self, name):
        field_map = {x: i for i, x in enumerate(self.components)}
        index = field_map.get(name)
        results = []
        for value in self.values:
            results.append(value[index])
        return results

    def get_max_utilization(self, specific_code: str = None):
        field_map = {x: i for i, x in enumerate(self.components) if x.startswith("uf")}
        max_uf = (0.0, None, None)
        for key, index in field_map.items():
            if specific_code is not None and key != specific_code:
                continue
            for res_point, value in enumerate(self.values):
                uf = value[index]
                if uf > max_uf[0]:
                    max_uf = (uf, key, res_point)
        return max_uf


@dataclass
class CCSesamResult:
    cc_file: str | pathlib.Path

    def __enter__(self):
        self._f = h5py.File(self.cc_file, "r")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._f.close()

    def get_all_results(self, name: str = None) -> dict[str, CCData]:
        cc_data = dict()
        for item in h5py_cc_iterator(self._f):
            comps = item.dtype.names
            for mem, data in groupby(item, key=lambda x: x["Member"]):
                mem_name = str(mem).split("member(")[-1].replace(")", "")[:-1]
                if name is not None and mem_name != name:
                    continue
                data = list(data)
                cc_data[mem_name] = CCData(mem_name, item.name, comps, data)

        return cc_data


def h5py_cc_iterator(g, prefix=""):
    for key in g.keys():
        item = g[key]
        path = "{}/{}".format(prefix, key)
        if isinstance(item, h5py.Dataset):
            yield item
        elif isinstance(item, h5py.Group):
            yield from h5py_cc_iterator(item, path)


def read_cc_file(cc_h5_file) -> dict[str, CCData]:
    with CCSesamResult(cc_h5_file) as cc_ses:
        results = cc_ses.get_all_results()

    return results



from __future__ import annotations

import json
import pathlib
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, List

import numpy as np

if TYPE_CHECKING:
    pass


@dataclass
class PlasticityModel:
    sig_p: List[float] = field(default=None)
    eps_p: List[float] = field(default=None)


class DnvGl16Mat(PlasticityModel):
    """
    This function returns a carbon steel Material object based on the formulas in DNVGL RP C208 (september 2016)


    :param t: Thickness of the material
    :param grade: Material Grade 'S355' or 'S420'
    :param mat_def: Material Definition
    :param eps_max: Maximum epsilon
    :param data_points: Number of datapoints
    :return: tuple of (eps_p, sig_p)
    """

    @staticmethod
    def dnv_thick_select(thick, grade):
        if thick < 0.016:
            mat_str = "t_16"
        elif 0.016 <= thick < 0.04:
            mat_str = "16_t_40"
        elif 0.04 <= thick < 0.063:
            mat_str = "40_t_63"
        else:
            if grade == "S420":
                mat_str = "40_t_63"
            else:
                mat_str = "63_t_100"
        return mat_str

    def __init__(self, t, grade, mat_def="Low", eps_max=0.3, data_points=200):
        super(DnvGl16Mat, self).__init__()
        self._mat_def = mat_def
        self._grade = grade
        self._t = t

        with open(pathlib.Path(__file__).parent / "resources/NLMatParams.json", "r") as f:
            params = json.load(f)

        thick_str = self.dnv_thick_select(t, grade)
        sig_prop = float(params[mat_def][thick_str]["q_prop"][grade])
        sig_yield_1 = float(params[mat_def][thick_str]["q_yield_1"][grade])
        sig_yield_2 = float(params[mat_def][thick_str]["q_yield_2"][grade])
        ep_y1 = float(params[mat_def][thick_str]["ep_y1"][grade])
        ep_y2 = float(params[mat_def][thick_str]["ep_y2"][grade])
        n = float(params[mat_def][thick_str]["n"][grade])
        K = float(params[mat_def][thick_str]["K"][grade])

        eps = np.linspace(ep_y2, eps_max, num=data_points)

        ep = list()
        ep.append(0.0)
        ep.append(ep_y1)
        ep.append(ep_y2)
        sig = list()
        sig.append(sig_prop)
        sig.append(sig_yield_1)
        sig.append(sig_yield_2)

        init_val = 0
        for e in eps:
            if init_val == 0:
                init_val = 1
            else:
                ep.append(e)
                sig.append(K * (e + (sig_yield_2 / K) ** (1.0 / n) - ep_y2) ** n)

        self.eps_p = ep
        self.sig_p = sig

    def __repr__(self):
        return f"Dnvgl16Mat({self.nl_name})"

    @property
    def nl_name(self):
        mat_prefix = "RP"
        mat_def_str = "L" if self._mat_def == "Low" else "M"
        return mat_prefix + "_" + self._grade + mat_def_str + "_" + self.dnv_thick_select(self._t, self._grade)

    @property
    def metadata(self):
        return dict(description="Carbon Steel nonlinear material based on DNVGL-RP-C208 (Sept. 2019)")



from __future__ import annotations

import re
from dataclasses import dataclass

import numpy as np

from ada.fem.formats.utils import get_ff_regex
from ada.sections.categories import BaseTypes

re_in = re.IGNORECASE | re.MULTILINE | re.DOTALL


@dataclass
class DataCard:
    name: str
    components: tuple

    def __post_init__(self):
        self._index_map = {x.upper(): i for i, x in enumerate(self.components)}

    def to_ff_re(self):
        return get_ff_regex(self.name, *self.components)

    def get_indices_from_names(self, names: list[str]) -> list[str] | str:
        res = [self._index_map[n.upper()] for n in names]
        return res if len(res) != 1 else res[0]

    def get_data_map_from_names(self, names: list[str], data: list) -> dict:
        indices = self.get_indices_from_names(names)
        output = dict()
        for i, name in zip(indices, names):
            output[name] = data[i]
        return output

    def cast_to_np(self, names, data):
        indices = self.get_indices_from_names(names)
        return np.array([[d[i] for i in indices] for d in data])

    @staticmethod
    def is_numeric(stripped: str):
        if stripped[0].isnumeric() is False and stripped[0] != "-":
            return False
        return True

    @staticmethod
    def str_to_proper_types(split_str: str) -> list[float | str]:
        return [float(x) if DataCard.is_numeric(x) else x for x in split_str.split()]


SEC_MAP = {
    "GIORH": (
        BaseTypes.IPROFILE,
        (("hz", "h"), ("ty", "t_w"), ("bt", "w_top"), ("tt", "t_ftop"), ("bb", "w_btn"), ("tb", "t_fbtn")),
    ),
    "GBOX": (
        BaseTypes.BOX,
        (("hz", "h"), ("ty", "t_w"), ("by", "w_top"), ("tt", "t_ftop"), ("by", "w_btn"), ("tb", "t_fbtn")),
    ),
}
MAT_MAP = {
    "MISOSEL": (
        ("young", "E"),
        ("poiss", "v"),
        ("rho", "rho"),
        ("damp", "zeta"),
        ("alpha", "alpha"),
        ("yield", "sig_y"),
    ),
    "MORSMEL": (("d11", "E"), ("ps1", "v"), ("rho", "rho"), ("alpha1", "alpha"), ("damp1", "zeta")),
}

# Nodes
GNODE = DataCard("GNODE", ("nodex", "nodeno", "ndof", "odof"))
GCOORD = DataCard("GCOORD", ("id", "x", "y", "z"))

# Elements
GELMNT1 = DataCard("GELMNT1", ("elnox", "elno", "eltyp", "eltyad", "nids"))
GELREF1 = DataCard(
    "GELREF1",
    (
        "elno",
        "matno",
        "addno",
        "intno",
        "mintno",
        "strano",
        "streno",
        "strepono",
        "geono",
        "fixno",
        "eccno",
        "transno",
        "members|",
    ),
)

# Beam Sections
# GIORH (I-section)
# GUSYI (unsymm.I-section)
# GCHAN (Channel section)
# GBOX (Box section)
# GPIPE (Pipe section)
# GBARM (Massive bar)
# GTONP (T on plate)
# GDOBO (Double box)
# GLSEC (L section)
# GIORHR
# GCHANR
# GLSECR


TDSECT = DataCard("TDSECT", ("nfield", "geono", "codnam", "codtxt", "set_name"))
re_sectnames = TDSECT.to_ff_re()
re_gbeamg = get_ff_regex(
    "GBEAMG",
    "geono",
    "comp",
    "area",
    "ix",
    "iy",
    "iz",
    "iyz",
    "wxmin",
    "wymin",
    "wzmin",
    "shary",
    "sharz",
    "shceny",
    "shcenz",
    "sy",
    "sz",
    "wy|",
    "wz|",
    "fabr|",
)
GIORH = DataCard("GIORH", ("geono", "hz", "ty", "bt", "tt", "bb", "tb", "sfy", "sfz", "NLOBYT|", "NLOBYB|", "NLOBZ|"))
GBOX = DataCard("GBOX", ("geono", "hz", "ty", "tb", "tt", "by", "sfy", "sfz"))
re_gbox = GBOX.to_ff_re()
re_gpipe = get_ff_regex("GPIPE", "geono", "di", "dy", "t", "sfy", "sfz")
re_gbarm = get_ff_regex("GBARM", "geono", "hz", "bt", "bb", "sfy", "sfz")

# Coordinate System
GUNIVEC = DataCard("GUNIVEC", ("transno", "unix", "uniy", "uniz"))

# Shell section
GELTH = DataCard("GELTH", ("geono", "th"))

# Other
re_bnbcd = get_ff_regex("BNBCD", "nodeno", "ndof", "content")
re_belfix = get_ff_regex(
    "BELFIX",
    "fixno",
    "opt",
    "trano",
    "unused",
    "a1|",
    "a2|",
    "a3|",
    "a4|",
    "a5|",
    "a6|",
)
re_mgsprng = get_ff_regex("MGSPRNG", "matno", "ndof", "bulk")
re_bnmass = get_ff_regex("BNMASS", "nodeno", "ndof", "m1", "m2", "m3", "m4", "m5", "m6")
re_mgmass = get_ff_regex("MGMASS", "matno", "ndof", "bulk")
re_geccen = get_ff_regex("GECCEN", "eccno", "ex", "ey", "ez")
re_bldep = get_ff_regex("BLDEP", "slave", "master", "nddof", "ndep", "bulk")

GSETMEMB = DataCard("GSETMEMB", ("nfield", "isref", "index", "istype", "isorig", "members"))
re_setmembs = GSETMEMB.to_ff_re()

TDSETNAM = DataCard("TDSETNAM", ("nfield", "isref", "codnam", "codtxt", "set_name"))
re_setnames = TDSETNAM.to_ff_re()

# Materials

TDMATER = DataCard("TDMATER", ("nfield", "geo_no", "codnam", "codtxt", "name"))
MISOSEL = DataCard("MISOSEL", ("matno", "young", "poiss", "rho", "damp", "alpha", "iyield", "yield"))
MORSMEL = DataCard(
    "MORSMEL",
    (
        "matno",
        "q1",
        "q2",
        "q3",
        "rho",
        "d11",
        "d21",
        "d22",
        "d31",
        "d32",
        "d33",
        "ps1",
        "ps2",
        "damp1",
        "damp2",
        "alpha1",
        "alpha2",
    ),
)

# Results
RDNODRES = DataCard("RDNODRES", ["nfield", "irdva", "lenrec"])
RVNODDIS = DataCard(
    "RVNODDIS", ["nfield", "ires", "inod", "irdva|", "itrans|", "U1|", "U2|", "U3|", "U4|", "U5|", "U6|"]
)
RDPOINTS = DataCard(
    "RDPOINTS", ["nfield", "ispalt", "iielno", "icoref", "ieltyp", "nsp", "ijkdim", "nsptra", "nlay", "bulk"]
)
RDSTRESS = DataCard("RDSTRESS", ["nfield", "irstrs", "lenrec", "bulk"])
RVSTRESS = DataCard("RVSTRESS", ["nfield", "ires", "iielno", "ispalt", "irstrs"])
RDIELCOR = DataCard("RDIELCOR", ["nfield, icoref", "igrid"])
RDRESREF = DataCard("RDRESREF", ["nfield", "ires", "irno", "ieres", "icalty", "complx", "numtyp", "bulk"])
RVFORCES = DataCard("RVFORCES", ["nfield", "ires", "ielno", "ispalt", "irforc|", "bulk|"])
RDFORCES = DataCard("RDFORCES", ["nfield", "irforc", "lenrec", "bulk|"])

re_rvforces = RVFORCES.to_ff_re()
re_rdforces = RDFORCES.to_ff_re()

re_rsumreac = get_ff_regex("RSUMREAC", "nfield", "ires", "ircomp", "x", "y", "z", "rx", "ry", "rz")
re_rvnodrea = get_ff_regex(
    "RVNODREA", "nfield", "ires", "inod", "irrea|", "irboc|", "itrans|", "F1|", "F2|", "F3|", "F4|", "F5|", "F6|"
)

re_rdrescmb = get_ff_regex("RDRESCMB", "nfield", "ires", "complx", "nres", "bulk")

TDRESREF = DataCard("TDRESREF", ("nfield", "ires", "codnam", "codtxt", "name"))
re_tdresref = TDRESREF.to_ff_re()
re_tdload = get_ff_regex("TDLOAD", "nfield", "llc", "codnam", "codtxt", "name")



from typing import TYPE_CHECKING

import numpy as np

from ada.config import Settings, get_logger
from ada.fem import Bc, Connector, ConnectorSection, Constraint, Elem, FemSet

if TYPE_CHECKING:
    from ada import FEM, Assembly

logger = get_logger()


def convert_ecc_to_mpc(fem: "FEM"):
    """Converts beam offsets to MPC constraints"""
    from ada import Node
    from ada.core.vector_utils import vector_length

    edited_nodes = dict()
    tol = Settings.point_tol

    def build_constraint(n_old, elem, ecc, i):
        n_new = edited_nodes[n_old.id]
        mat = np.eye(3)
        new_p = np.dot(mat, ecc) + n_old.p
        n_new_ = Node(new_p, parent=elem.parent)
        if vector_length(n_new_.p - n_new.p) > tol:
            elem.parent.nodes.add(n_new_, allow_coincident=True)
            m_set = FemSet(f"el{elem.id}_mpc{i + 1}_m", [n_new_], "nset")
            s_set = FemSet(f"el{elem.id}_mpc{i + 1}_s", [n_old], "nset")
            c = Constraint(
                f"el{elem.id}_mpc{i + 1}_co",
                Constraint.TYPES.MPC,
                m_set,
                s_set,
                mpc_type="Beam",
                parent=elem.parent,
            )
            elem.parent.add_constraint(c)
            elem.nodes[i] = n_new_
            edited_nodes[n_old.id] = n_new_

        else:
            elem.nodes[i] = n_new
            edited_nodes[n_old.id] = n_new

    def build_mpc_for_end(elem, n_old, ecc, i):
        if n_old.id in edited_nodes.keys():
            build_constraint(n_old, elem, ecc, i)
        else:
            mat = np.eye(3)
            new_p = np.dot(mat, ecc) + n_old.p
            n_new = Node(new_p, parent=elem.parent)
            elem.parent.nodes.add(n_new, allow_coincident=True)
            m_set = FemSet(f"el{elem.id}_mpc{i + 1}_m", [n_new], "nset")
            s_set = FemSet(f"el{elem.id}_mpc{i + 1}_s", [n_old], "nset")
            c = Constraint(
                f"el{elem.id}_mpc{i + 1}_co",
                Constraint.TYPES.MPC,
                m_set,
                s_set,
                mpc_type="Beam",
                parent=elem.parent,
            )
            elem.parent.add_constraint(c)
            elem.nodes[i] = n_new
            edited_nodes[n_old.id] = n_new

    def build_mpc(elem: Elem):
        if elem.eccentricity.end1 is not None:
            n_old = elem.eccentricity.end1.node
            ecc = elem.eccentricity.end1.ecc_vector
            i = elem.nodes.index(n_old)
            build_mpc_for_end(elem, n_old, ecc, i)
        if elem.eccentricity.end2 is not None:
            n_old = elem.eccentricity.end2.node
            ecc = elem.eccentricity.end2.ecc_vector
            i = elem.nodes.index(n_old)
            build_mpc_for_end(elem, n_old, ecc, i)

    [build_mpc(el) for el in fem.elements.lines_ecc]


def convert_hinges_2_couplings(fem: "FEM"):
    """Convert beam hinges to coupling constraints"""
    from ada import Node
    from ada.core.utils import Counter
    from ada.fem.elements import Hinge

    constrain_ids = []

    max_node_id = fem.nodes.max_nid
    new_node_id = Counter(int(max_node_id + 10000))

    def convert_hinge(elem: Elem, hinge: Hinge):
        if hinge.constraint_ref is not None:
            return
        n = hinge.fem_node
        csys = hinge.csys
        d = hinge.retained_dofs

        n2 = Node(n.p, next(new_node_id), parent=elem.parent)
        elem.parent.nodes.add(n2, allow_coincident=True)
        i = elem.nodes.index(n)
        elem.nodes[i] = n2

        if elem.eccentricity is not None:
            if elem.eccentricity.end1 is not None:
                if n == elem.eccentricity.end1.node:
                    elem.eccentricity.end1.node = n2

            if elem.eccentricity.end2 is not None:
                if n == elem.eccentricity.end2.node:
                    elem.eccentricity.end2.node = n2

        if n2.id not in constrain_ids:
            constrain_ids.append(n2.id)
        else:
            logger.error(f"Hinged node {n2} cannot be added twice to different couplings")
            return None

        m_set = FemSet(f"el{elem.id}_hinge{i + 1}_m", [n], "nset")
        s_set = FemSet(f"el{elem.id}_hinge{i + 1}_s", [n2], "nset")

        elem.parent.add_set(m_set)
        elem.parent.add_set(s_set)
        c = Constraint(
            f"el{elem.id}_hinge{i + 1}_co",
            Constraint.TYPES.COUPLING,
            m_set,
            s_set,
            d,
            csys=csys,
        )
        elem.parent.add_constraint(c)
        hinge.constraint_ref = c
        logger.info(f"added constraint {c}")

    for el in fem.elements.lines_hinged:
        if el.hinge_prop.end1 is not None:
            convert_hinge(el, el.hinge_prop.end1)
        if el.hinge_prop.end2 is not None:
            convert_hinge(el, el.hinge_prop.end2)


def convert_springs_to_connectors(assembly: "Assembly"):
    """Converts all single noded springs to connector elements"""
    from ada import Node

    for p in assembly.get_all_subparts():
        for spring in p.fem.springs.values():
            n1 = spring.nodes[0]
            n2 = Node(n1.p - np.array([0, 0, 10e-3]))
            assembly.fem.add_rp(spring.name + "_rp", n2)
            fs = FemSet(spring.name + "_bc", [n2], "nset")
            assembly.fem.add_set(fs)
            assembly.fem.add_bc(Bc(spring.name + "_bc", fs, [1, 2, 3, 4, 5, 6]))
            diag = []
            for dof, row in enumerate(spring.stiff):
                for j, stiffness in enumerate(row):
                    if dof == j:
                        diag.append(stiffness)

            con_sec = ConnectorSection(spring.name + "_consec", diag, [])
            assembly.fem.add_connector_section(con_sec)
            con = Connector(spring.name + "_con", spring.id, n1, n2, "bushing", con_sec)
            assembly.fem.add_connector(con)
        p.fem._springs = dict()
        p.fem.elements.filter_elements(delete_elem=["SPRING1"])



from typing import TYPE_CHECKING

from ada.concepts.containers import Nodes
from ada.concepts.points import Node
from ada.fem.formats.utils import str_to_int

from . import cards

if TYPE_CHECKING:
    from ada.fem import FEM


def get_nodes(bulk_str: str, parent: "FEM") -> Nodes:
    nodes = [get_node(m, parent) for m in cards.GCOORD.to_ff_re().finditer(bulk_str)]
    return Nodes(nodes, parent=parent)


def renumber_nodes(bulk_str: str, fem: "FEM") -> None:
    node_map = {nodeno: nodex for nodex, nodeno in map(get_nodeno, cards.GNODE.to_ff_re().finditer(bulk_str))}
    fem.nodes.renumber(renumber_map=node_map)


def get_nodeno(m_gnod):
    d = m_gnod.groupdict()
    return str_to_int(d["nodex"]), str_to_int(d["nodeno"])


def get_node(m, parent):
    d = m.groupdict()
    return Node([float(d["x"]), float(d["y"]), float(d["z"])], str_to_int(d["id"]), parent=parent)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ada.fem.steps import FieldOutput, Step


def _create_field_output(field: FieldOutput, nl_geom) -> str:
    # TODO: Make this a function of field output. Not randomly chosen hardcoded variables
    _ = {"S": "SIGM_ELNO"}

    default_contrainte = ["EFGE_ELNO", "EFGE_NOEU", "SIGM_ELNO", "SIEF_ELNO", "SIPO_ELNO", "SIPM_ELNO"]
    default_deformation = ["EPSI_ELNO", "EPSP_ELNO"]
    default_force = ["REAC_NODA"]

    defaults_def_str = ""
    defaults_c_str = ",".join([f'"{x}"' for x in default_contrainte])

    if nl_geom:
        defaults_def_str = "\n        DEFORMATION=(" + ",".join([f'"{x}"' for x in default_deformation]) + "),"

    defaults_forc_str = ",".join([f'"{x}"' for x in default_force])

    return f"""
result = CALC_CHAMP(
        reuse=result, RESULTAT=result,
        CONTRAINTE=({defaults_c_str}),
        FORCE= ({defaults_forc_str}),{defaults_def_str}
)\n"""


def _create_post_calc_nl_geom(step):
    return """stress = POST_CHAMP(
    EXTR_COQUE=_F(
        NIVE_COUCHE='MOY',
        NOM_CHAM=('SIGM_ELNO', ),
        NUME_COUCHE=1
    ),
    RESULTAT=result
)

stress = CALC_CHAMP(
    reuse=stress,
    CONTRAINTE=('SIGM_NOEU', ),
    RESULTAT=stress
)

strain = POST_CHAMP(
    EXTR_COQUE=_F(
        NIVE_COUCHE='MOY',
    NOM_CHAM=('EPSI_ELNO', ),
    NUME_COUCHE=1),
    RESULTAT=result
)

strainP = POST_CHAMP(
    EXTR_COQUE=_F(
        NIVE_COUCHE='MOY',
    NOM_CHAM=('EPSP_ELNO', ),
    NUME_COUCHE=1),
    RESULTAT=result
)"""


def final_writer():
    return """
    IMPR_RESU(
    RESU=(
        _F(
            NOM_CHAM=("DEPL", "EFGE_ELNO", "EFGE_NOEU"),
            NOM_CHAM_MED=("DISP", "GEN_FORCES_ELEM", "GEN_FORCES_NODES"),
            RESULTAT=result,
        ),
        _F(
            NOM_CHAM=("SIGM_ELNO", "SIGM_NOEU"),
            NOM_CHAM_MED=("STRESSES_ELEM", "STRESSES_NODES"),
            RESULTAT=stress,
        ),
        _F(
            NOM_CHAM=("EPSI_ELNO",),
            NOM_CHAM_MED=("STRAINS_ELEM",),
            RESULTAT=strain,
        ),
        _F(
            NOM_CHAM=("EPSP_ELNO",),
            NOM_CHAM_MED=("PLASTIC_STRAINS_ELEM",),
            RESULTAT=strainP,
        ),
    ),
    UNITE=80,
)
"""


def create_field_output_str(step: Step) -> str:
    out_str = ""
    for f in step.field_outputs:
        out_str += _create_field_output(f, step.nl_geom)

    return out_str



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Callable, Iterable, List, Optional, Union

import numpy as np

from ada.base.physical_objects import BackendGeom
from ada.base.units import Units
from ada.concepts.bounding_box import BoundingBox
from ada.concepts.curves import CurvePoly, CurveRevolve
from ada.concepts.points import Node, get_singular_node_by_volume
from ada.concepts.transforms import Placement
from ada.config import Settings, get_logger
from ada.core.utils import Counter, roundoff
from ada.core.vector_utils import (
    angle_between,
    calc_yvec,
    calc_zvec,
    is_between_endpoints,
    is_parallel,
    unit_vector,
    vector_length,
)
from ada.materials import Material
from ada.materials.utils import get_material
from ada.sections import Section
from ada.sections.utils import get_section

if TYPE_CHECKING:
    from OCC.Core.TopoDS import TopoDS_Shape

    from ada.concepts.connections import JointBase
    from ada.concepts.spatial import Part
    from ada.fem.elements import HingeProp
    from ada.ifc.store import IfcStore

section_counter = Counter(1)
material_counter = Counter(1)
logger = get_logger()


class Justification(Enum):
    NA = "neutral axis"
    TOS = "top of steel"


class Beam(BackendGeom):
    """
    The base Beam object

    :param n1: Start position of beam. List or Node object
    :param n2: End position of beam. List or Node object
    :param sec: Section definition. Str or Section Object
    :param mat: Material. Str or Material object. String: ['S355' & 'S420'] (default is 'S355' if None is parsed)
    :param name: Name of beam
    :param tap: Tapering of beam. Str or Section object
    :param jusl: Justification of Beam centreline
    :param curve: Curve
    """

    JUSL_TYPES = Justification

    def __init__(
        self,
        name,
        n1: Node | Iterable = None,
        n2: Node | Iterable = None,
        sec: str | Section = None,
        mat: str | Material = None,
        tap: str | Section = None,
        jusl=JUSL_TYPES.NA,
        up=None,
        angle=0.0,
        curve: CurvePoly | CurveRevolve = None,
        e1=None,
        e2=None,
        colour=None,
        parent: Part = None,
        metadata=None,
        opacity=1.0,
        units=Units.M,
        guid=None,
        placement=Placement(),
        ifc_store: IfcStore = None,
    ):
        super().__init__(
            name,
            metadata=metadata,
            units=units,
            guid=guid,
            placement=placement,
            ifc_store=ifc_store,
            colour=colour,
            opacity=opacity,
        )
        if curve is not None:
            curve.parent = self
            if type(curve) is CurvePoly:
                n1 = curve.points3d[0]
                n2 = curve.points3d[-1]
            elif type(curve) is CurveRevolve:
                n1 = curve.p1
                n2 = curve.p2
            else:
                raise ValueError(f'Unsupported curve type "{type(curve)}"')

        self._curve = curve
        self._n1 = n1 if type(n1) is Node else Node(n1[:3], units=units)
        self._n2 = n2 if type(n2) is Node else Node(n2[:3], units=units)
        self._jusl = jusl

        self._connected_to = []
        self._connected_end1 = None
        self._connected_end2 = None
        self._tos = None
        self._e1 = e1
        self._e2 = e2
        self._hinge_prop = None

        self._parent = parent
        self._bbox = None

        # Section and Material setup
        self._section, self._taper = get_section(sec)
        self._section.refs.append(self)

        self._material = get_material(mat)
        self._material.refs.append(self)

        if tap is not None:
            self._taper, _ = get_section(tap)

        self._taper.refs.append(self)

        self._section.parent = self
        self._taper.parent = self

        # Define orientations
        self._init_orientation(angle, up)
        self.add_beam_to_node_refs()

    @staticmethod
    def from_list_of_coords(
        list_of_coords: list[tuple], sec: Section | str, mat: Material | str = None, name_gen: Callable = None
    ) -> list[Beam]:
        beams = []
        ngen = name_gen if name_gen is not None else Counter(prefix="bm")
        for p1, p2 in zip(list_of_coords[:-1], list_of_coords[1:]):
            beams.append(Beam(next(ngen), p1, p2, sec, mat))
        return beams

    def _init_orientation(self, angle=None, up=None) -> None:
        xvec = unit_vector(self.n2.p - self.n1.p)
        tol = 1e-3
        zvec = calc_zvec(xvec)
        gup = np.array(zvec)

        if up is None:
            if angle != 0.0 and angle is not None:
                from pyquaternion import Quaternion

                my_quaternion = Quaternion(axis=xvec, degrees=angle)
                rot_mat = my_quaternion.rotation_matrix
                up = np.array([roundoff(x) if abs(x) != 0.0 else 0.0 for x in np.matmul(gup, np.transpose(rot_mat))])
            else:
                up = np.array([roundoff(x) if abs(x) != 0.0 else 0.0 for x in gup])
            yvec = calc_yvec(xvec, up)
        else:
            if (len(up) == 3) is False:
                raise ValueError("Up vector must be length 3")
            if vector_length(xvec - up) < tol:
                raise ValueError("The assigned up vector is too close to your beam direction")
            yvec = calc_yvec(xvec, up)
            # TODO: Fix improper calculation of angle (e.g. xvec = [1,0,0] and up = [0,1,0] should be 270?
            rad = angle_between(up, zvec)
            angle = np.rad2deg(rad)
            up = np.array(up)

        # lup = np.cross(xvec, yvec)
        self._xvec = xvec
        self._yvec = np.array([roundoff(x) for x in yvec])
        self._up = up
        self._angle = angle

    def is_point_on_beam(self, point: Union[np.ndarray, Node]) -> bool:
        if isinstance(point, Node):
            point = point.p

        return is_between_endpoints(point, self.n1.p, self.n2.p, incl_endpoints=True)

    def split_beam(
        self, point: Union[Node, np.ndarray] = None, fraction: float = None, length: float = None
    ) -> Optional[Beam]:
        """
        Split beam into two parts, and returns the new beam. Prioritizes input arguments in given order if  given
        multiple input.

        :param point:
        :param fraction: Fraction of the beam length from Node n1.
        :param length: Length of the beam from Node n1.
        """

        if isinstance(point, Node):
            point = point.p

        if point is not None:
            splitting_node = self.get_node_on_beam_by_point(point)
        elif fraction is not None:
            splitting_node = self.get_node_on_beam_by_fraction(fraction)
        elif length is not None:
            length_fraction = length / self.length
            splitting_node = self.get_node_on_beam_by_fraction(length_fraction)
        else:
            logger.warning(f"Beam {self} is not split as inconclusive info is provided.")
            return None

        node_on_beam = self.parent.fem.nodes.add(splitting_node)
        splitted_beam = self.get_split_beam(node_on_beam)
        return splitted_beam

    def get_node_on_beam_by_point(self, point: np.ndarray) -> Node:
        """Returns node on beam from point"""
        if not is_between_endpoints(point, self.n1.p, self.n2.p):
            raise ValueError(f"The node is not on line and between the beam end points, p: {point}, bm: {self}")

        return get_singular_node_by_volume(self.parent.fem.nodes, point)

    def get_node_on_beam_by_fraction(self, fraction: float) -> Node:
        """Returns node as a fraction of the beam length from n1-node."""

        if not 0.0 < fraction < 1.0:
            raise ValueError(f"Fraction {fraction} is not between 0 and 1")

        return get_singular_node_by_volume(self.parent.fem.nodes, self.n1.p + fraction * self.length * self.xvec)

    def get_split_beam(self, node: Node, section: Section = None, material: Material = None) -> Beam:
        """Returns new beam. Setting splitting node to n2-node on self and to n1-node on the new beam."""

        new_beam = Beam(
            name=f"{self.name}_2",
            n1=node,
            n2=self.n2,
            sec=self.section if section is None else section,
            mat=self.material if material is None else material,
            tap=self.taper,
            jusl=self.jusl,
            up=self.up,
            e1=self.e1,
            e2=self.e2,
            colour=self.colour,
            parent=self.parent,
            metadata=self.metadata,
            opacity=self.opacity,
            units=self.units,
        )

        self.name = f"{self.name}_1"
        self.n2 = node
        return new_beam

    def updating_nodes(self, old_node: Node, new_node: Node) -> None:
        """Exchanging node on beam"""
        if old_node is self.n1:
            self.n1 = new_node
        elif old_node is self.n2:
            self.n2 = new_node
        else:
            raise NodeNotOnEndpointError(f"{old_node} is on either endpoint: {self.nodes}")

    def get_outer_points(self):
        from itertools import chain

        from ada.core.vector_utils import local_2_global_points

        section_profile = self.section.get_section_profile(False)
        if section_profile.disconnected:
            ot = list(chain.from_iterable([x.points2d for x in section_profile.outer_curve_disconnected]))
        else:
            ot = section_profile.outer_curve.points2d

        yv = self.yvec
        xv = self.xvec
        p1 = self.n1.p
        p2 = self.n2.p

        nodes_p1 = local_2_global_points(ot, p1, yv, xv)
        nodes_p2 = local_2_global_points(ot, p2, yv, xv)

        return nodes_p1, nodes_p2

    def calc_con_points(self, point_tol=Settings.point_tol):
        from ada.core.vector_utils import sort_points_by_dist

        a = self.n1.p
        b = self.n2.p
        points = [tuple(con.centre) for con in self.connected_to]

        def is_mem_eccentric(mem, centre):
            is_ecc = False
            end = None
            if point_tol < vector_length(mem.n1.p - centre) < mem.length * 0.9:
                is_ecc = True
                end = mem.n1.p
            if point_tol < vector_length(mem.n2.p - centre) < mem.length * 0.9:
                is_ecc = True
                end = mem.n2.p
            return is_ecc, end

        if len(self.connected_to) == 1:
            con = self.connected_to[0]
            if con.main_mem == self:
                for m in con.beams:
                    if m != self:
                        is_ecc, end = is_mem_eccentric(m, con.centre)
                        if is_ecc:
                            logger.info(f'do something with end "{end}"')
                            points.append(tuple(end))

        midpoints = []
        prev_p = None
        for p in sort_points_by_dist(a, points):
            p = np.array(p)
            bmlen = self.length
            vlena = vector_length(p - a)
            vlenb = vector_length(p - b)

            if prev_p is not None:
                if vector_length(p - prev_p) < point_tol:
                    continue

            if vlena < point_tol:
                self._connected_end1 = self.connected_to[points.index(tuple(p))]
                prev_p = p
                continue

            if vlenb < point_tol:
                self._connected_end2 = self.connected_to[points.index(tuple(p))]
                prev_p = p
                continue

            if vlena > bmlen or vlenb > bmlen:
                prev_p = p
                continue

            midpoints += [p]
            prev_p = p

        return midpoints

    def copy_to(self, p1, p2, name: str) -> Beam:
        return Beam(name, p1, p2, sec=self.section, tap=self.taper, mat=self.material)

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if isinstance(value, str):
            value = Units.from_str(value)
        if self._units == value:
            return

        self.n1.units = value
        self.n2.units = value
        self.section.units = value
        self.material.units = value
        for pen in self.penetrations:
            pen.units = value
        self._units = value

    @property
    def section(self) -> Section:
        return self._section

    @section.setter
    def section(self, value: Section):
        self._section = value

    @property
    def taper(self) -> Section:
        return self._taper

    @taper.setter
    def taper(self, value: Section):
        self._taper = value

    @property
    def material(self) -> Material:
        return self._material

    @material.setter
    def material(self, value: Material):
        self._material = value

    @property
    def member_type(self):
        from ada.core.vector_utils import is_parallel

        xvec = self.xvec
        if is_parallel(xvec, [0.0, 0.0, 1.0], tol=1e-1):
            mtype = "Column"
        elif xvec[2] == 0.0:
            mtype = "Girder"
        else:
            mtype = "Brace"

        return mtype

    @property
    def connected_to(self) -> List[JointBase]:
        return self._connected_to

    @property
    def connected_end1(self):
        return self._connected_end1

    @property
    def connected_end2(self):
        return self._connected_end2

    @property
    def length(self) -> float:
        """Returns the length of the beam"""
        p1 = self.n1.p.copy()
        p2 = self.n2.p.copy()

        if self.e1 is not None:
            p1 += self.e1
        if self.e2 is not None:
            p2 += self.e2
        return vector_length(p2 - p1)

    @property
    def jusl(self):
        """Justification line"""
        return self._jusl

    @property
    def ori(self):
        """Get the x-vector, y-vector and z-vector of a given beam"""

        return self.xvec, self.yvec, self.up

    @property
    def xvec(self) -> np.ndarray:
        """Local X-vector"""
        return self._xvec

    @property
    def yvec(self) -> np.ndarray:
        """Local Y-vector"""
        return self._yvec

    @property
    def up(self) -> np.ndarray:
        return self._up

    @property
    def xvec_e(self) -> np.ndarray:
        """Local X-vector (including eccentricities)"""
        if self.e1 is not None:
            p1 = np.array([float(x) + float(self.e1[i]) for i, x in enumerate(self.n1.p.copy())])
        else:
            p1 = self.n1.p.copy()
        if self.e2 is not None:
            p2 = np.array([float(x) + float(self.e2[i]) for i, x in enumerate(self.n2.p.copy())])
        else:
            p2 = self.n2.p.copy()
        return unit_vector(p2 - p1)

    @property
    def n1(self) -> Node:
        return self._n1

    @n1.setter
    def n1(self, new_node: Node):
        self._n1.remove_obj_from_refs(self)
        self._n1 = new_node  # .get_main_node_at_point()
        self._n1.add_obj_to_refs(self)

    @property
    def n2(self) -> Node:
        return self._n2

    @n2.setter
    def n2(self, new_node: Node):
        self._n2.remove_obj_from_refs(self)
        self._n2 = new_node  # .get_main_node_at_point()
        self._n2.add_obj_to_refs(self)

    def bbox(self) -> BoundingBox:
        """Bounding Box of beam"""
        if self._bbox is None:
            self._bbox = BoundingBox(self)

        return self._bbox

    @property
    def e1(self) -> np.ndarray:
        return self._e1

    @e1.setter
    def e1(self, value):
        self._e1 = np.array(value)

    @property
    def e2(self) -> np.ndarray:
        return self._e2

    @e2.setter
    def e2(self, value):
        self._e2 = np.array(value)

    @property
    def hinge_prop(self) -> HingeProp:
        return self._hinge_prop

    @hinge_prop.setter
    def hinge_prop(self, value: HingeProp):
        value.beam_ref = self
        if value.end1 is not None:
            value.end1.concept_node = self.n1
        if value.end2 is not None:
            value.end2.concept_node = self.n2
        self._hinge_prop = value

    @property
    def curve(self) -> CurvePoly:
        return self._curve

    def line(self):
        from ada.occ.utils import make_wire_from_points

        # midpoints = self.calc_con_points()
        # points = [self.n1.p]
        # points += midpoints
        # points += [self.n2.p]

        points = [self.n1.p, self.n2.p]

        return make_wire_from_points(points)

    def shell(self) -> TopoDS_Shape:
        from ada.occ.utils import apply_penetrations, create_beam_geom

        geom = apply_penetrations(create_beam_geom(self, False), self.penetrations)

        return geom

    def solid(self) -> TopoDS_Shape:
        from ada.occ.utils import apply_penetrations, create_beam_geom

        geom = apply_penetrations(create_beam_geom(self, True), self.penetrations)

        return geom

    @property
    def nodes(self) -> list[Node]:
        return [self.n1, self.n2]

    @property
    def angle(self) -> float:
        return self._angle

    @angle.setter
    def angle(self, value: float):
        self._init_orientation(value)

    @property
    def vector(self) -> np.ndarray:
        """Returns the full length beam vector"""
        return self.length * self.xvec

    def is_on_beam(self, point: Node) -> bool:
        """Returns if a node is on the beam axis including endpoints"""
        return point in self.nodes or is_between_endpoints(point.p, self.n1.p, self.n2.p)

    def add_beam_to_node_refs(self) -> None:
        """Add beam to refs on nodes"""
        for beam_node in self.nodes:
            beam_node.add_obj_to_refs(self)

    def remove_beam_from_node_refs(self) -> None:
        """Remove beam from refs on nodes"""
        for beam_node in self.nodes:
            beam_node.remove_obj_from_refs(self)

    def is_equivalent(self, item: Beam) -> bool:
        """Returns equivalent beam-type, meaning beam characteristics are the same but NOT the same beam"""
        return (self.section, self.material) == (item.section, item.material) and self != item

    def get_beam_extensions(self) -> Iterable[Beam]:
        """Returns connected beams with same material and section at beam end-nodes, that are parallel"""

        def is_equal_beamtype(item) -> bool:
            return isinstance(item, Beam) and self.is_equivalent(item) and is_parallel(self.xvec, item.xvec)

        return list(filter(is_equal_beamtype, self.n1.refs + self.n2.refs))

    def is_weak_axis_stiffened(self, other_beam: Beam) -> bool:
        """Assumes rotation local z-vector (up) is weak axis"""
        return np.abs(np.dot(self.up, other_beam.xvec)) < Settings.point_tol and self is not other_beam

    def is_strong_axis_stiffened(self, other_beam: Beam) -> bool:
        """Assumes rotation local y-vector is strong axis"""
        return np.abs(np.dot(self.yvec, other_beam.xvec)) < Settings.point_tol and self is not other_beam

    def __hash__(self):
        return hash(self.guid)

    def __eq__(self, other: Beam):
        for key, val in self.__dict__.items():
            if "parent" in key or key in ["_ifc_settings", "_ifc_elem"]:
                continue
            oval = other.__dict__[key]

            if type(val) in (list, tuple, np.ndarray):
                if False in [x == y for x, y in zip(oval, val)]:
                    return False
            try:
                res = oval != val
            except ValueError as e:
                logger.error(e)
                return True

            if res is True:
                return False

        return True

    def __repr__(self):
        p1s = self.n1.p.tolist()
        p2s = self.n2.p.tolist()
        secn = self.section.sec_str
        matn = self.material.name
        return f'Beam("{self.name}", {p1s}, {p2s}, "{secn}", "{matn}")'

    def __setstate__(self, state):
        self.__dict__ = state

    def __getstate__(self):
        return self.__dict__


class NodeNotOnEndpointError(Exception):
    pass



from ada.fem.shapes.definitions import LineShapes, ShellShapes, SolidShapes

gmsh_map = {
    "Triangle 3": ShellShapes.TRI,
    "Triangle 6": ShellShapes.TRI6,
    "Quadrilateral 4": ShellShapes.QUAD,
    "Quadrilateral 8": ShellShapes.QUAD8,
    "Line 2": LineShapes.LINE,
    "Line 3": LineShapes.LINE3,
    "Tetrahedron 4": SolidShapes.TETRA,
    "Tetrahedron 10": SolidShapes.TETRA10,
    "Hexahedron 8": SolidShapes.HEX8,
    "Hexahedron 20": SolidShapes.HEX20,
}



from __future__ import annotations

from typing import TYPE_CHECKING

from ada.fem import Constraint, FemSet, Surface

from .helper_utils import get_instance_name
from .write_orientations import csys_str
from .write_surfaces import surface_str

if TYPE_CHECKING:
    from ada import FEM

# Coupling definition:
# https://abaqus-docs.mit.edu/2017/English/SIMACAEKEYRefMap/simakey-r-coupling.htm#simakey-r-coupling


def constraints_str(fem: FEM, written_on_assembly_level: bool):
    if len(fem.constraints.keys()) == 0:
        return "** No Constraints"

    return "\n".join([constraint_str(c, written_on_assembly_level) for c in fem.constraints.values()])


def constraint_str(constraint: Constraint, on_assembly_level: bool):
    if constraint.type == Constraint.TYPES.COUPLING:
        return _coupling(constraint, on_assembly_level)
    elif constraint.type == Constraint.TYPES.TIE:
        return _tie(constraint, on_assembly_level)
    elif constraint.type == Constraint.TYPES.RIGID_BODY:
        rnode = get_instance_name(constraint.m_set, on_assembly_level)
        return f"*Rigid Body, ref node={rnode}, elset={get_instance_name(constraint.s_set, on_assembly_level)}"
    elif constraint.type == Constraint.TYPES.MPC:
        return _mpc(constraint, on_assembly_level)
    elif constraint.type == Constraint.TYPES.SHELL2SOLID:
        return _shell2solid(constraint, on_assembly_level)
    else:
        raise NotImplementedError(f"{constraint.type}")


def _coupling(constraint: Constraint, on_assembly_level: bool):
    dofs_str = "".join([f" {x[0]}, {x[1]}\n" if type(x) != int else f" {x}, {x}\n" for x in constraint.dofs]).rstrip()

    if type(constraint.s_set) is FemSet:
        new_surf = surface_str(
            Surface(
                f"{constraint.name}_surf",
                Surface.TYPES.NODE,
                constraint.s_set,
                1.0,
                parent=constraint.s_set.parent,
            ),
            on_assembly_level,
        )
        surface_ref = f"{constraint.name}_surf"
        add_str = new_surf
    else:
        add_str = "**"
        surface_ref = get_instance_name(constraint.s_set, on_assembly_level)

    if constraint.csys is not None:
        new_csys_str = "\n" + csys_str(constraint.csys, on_assembly_level)
        cstr = f", Orientation={constraint.csys.name.upper()}"
    else:
        cstr = ""
        new_csys_str = ""

    rnode = f"{get_instance_name(constraint.m_set.members[0], on_assembly_level)}"
    return f"""** ----------------------------------------------------------------
** Coupling element {constraint.name}
** ----------------------------------------------------------------{new_csys_str}
** COUPLING {constraint.name}
{add_str}
*COUPLING, CONSTRAINT NAME={constraint.name}, REF NODE={rnode}, SURFACE={surface_ref}{cstr}
*KINEMATIC
{dofs_str}""".rstrip()


def _mpc(constraint, on_assembly_level: bool):
    mpc_type = constraint.mpc_type
    m_members = constraint.m_set.members
    s_members = constraint.s_set.members
    mpc_vars = "\n".join(
        [
            f" {mpc_type},{get_instance_name(m, on_assembly_level):>8},{get_instance_name(s, on_assembly_level):>8}"
            for m, s in zip(m_members, s_members)
        ]
    )
    return f"** Constraint: {constraint.name}\n*MPC\n{mpc_vars}"


def _shell2solid(constraint, on_assembly_level: bool):
    mname = constraint.m_set.name
    sname = constraint.s_set.name
    influence = constraint.influence_distance
    influence_str = "" if influence is None else f", influence distance={influence}"
    return (
        f"** Constraint: {constraint.name}\n*Shell to Solid Coupling, "
        f"constraint name={constraint.name}{influence_str}\n{mname}, {sname}"
    )


def _tie(constraint: Constraint, on_assembly_level: bool) -> str:
    num = 80
    pos_tol_str = ""
    if constraint.pos_tol is not None:
        pos_tol_str = f", position tolerance={constraint.pos_tol},"

    coupl_text = "**" + num * "-" + """\n** COUPLING {}\n""".format(constraint.name) + "**" + num * "-" + "\n"
    name = constraint.name

    adjust = constraint.metadata.get("adjust", "no")

    coupl_text += f"""** Constraint: {name}
*Tie, name={name}, adjust={adjust}{pos_tol_str}
{constraint.m_set.name}, {constraint.s_set.name}"""
    return coupl_text
